import copy
import importlib
import sys

from unittest import mock

import unittest

import rollbar
from rollbar.lib._async import AsyncMock
from rollbar.test import BaseTest

ALLOWED_PYTHON_VERSION = sys.version_info >= (3, 5)
ASYNC_REPORT_ENABLED = sys.version_info >= (3, 6)


@unittest.skipUnless(ALLOWED_PYTHON_VERSION, 'ASGI implementation requires Python3.5+')
class ReporterMiddlewareTest(BaseTest):
    default_settings = copy.deepcopy(rollbar.SETTINGS)

    def setUp(self):
        importlib.reload(rollbar)
        rollbar.SETTINGS = copy.deepcopy(self.default_settings)
        rollbar.SETTINGS['handler'] = 'async'

    @mock.patch('rollbar.report_exc_info')
    def test_should_catch_and_report_errors(self, mock_report):
        from rollbar.contrib.asgi.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        testapp = ReporterMiddleware(FailingTestASGIApp())

        with self.assertRaises(RuntimeError):
            run(testapp({'type': 'http'}, None, None))

        self.assertTrue(mock_report.called)

        args, kwargs = mock_report.call_args
        self.assertEqual(kwargs, {})

        exc_type, exc_value, exc_tb = args[0]

        self.assertEqual(exc_type, RuntimeError)
        self.assertIsInstance(exc_value, RuntimeError)

    @mock.patch('rollbar._check_config', return_value=True)
    @mock.patch('rollbar.send_payload')
    def test_should_add_framework_name_to_payload(self, mock_send_payload, *mocks):
        import rollbar
        from rollbar.contrib.asgi.middleware import ReporterMiddleware

        self.assertIsNone(rollbar.BASE_DATA_HOOK)

        ReporterMiddleware(None)  # invoke integration
        rollbar.report_exc_info()

        self.assertTrue(mock_send_payload.called)
        payload = mock_send_payload.call_args[0][0]

        self.assertIn('asgi', payload['data']['framework'])

    @unittest.skipUnless(ASYNC_REPORT_ENABLED, 'Requires Python 3.6+')
    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_async_report_exc_info_if_default_handler(
        self, sync_report_exc_info, async_report_exc_info
    ):
        import rollbar
        from rollbar.contrib.asgi.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        rollbar.SETTINGS['handler'] = 'default'
        testapp = ReporterMiddleware(FailingTestASGIApp())

        with self.assertRaises(RuntimeError):
            run(testapp({'type': 'http'}, None, None))

        self.assertTrue(async_report_exc_info.called)
        self.assertFalse(sync_report_exc_info.called)

    @unittest.skipUnless(ASYNC_REPORT_ENABLED, 'Requires Python 3.6+')
    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_async_report_exc_info_if_any_async_handler(
        self, sync_report_exc_info, async_report_exc_info
    ):
        import rollbar
        from rollbar.contrib.asgi.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        rollbar.SETTINGS['handler'] = 'httpx'
        testapp = ReporterMiddleware(FailingTestASGIApp())

        with self.assertRaises(RuntimeError):
            run(testapp({'type': 'http'}, None, None))

        self.assertTrue(async_report_exc_info.called)
        self.assertFalse(sync_report_exc_info.called)

    @unittest.skipUnless(ASYNC_REPORT_ENABLED, 'Requires Python 3.6+')
    @mock.patch('logging.Logger.warning')
    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_sync_report_exc_info_if_non_async_handlers(
        self, sync_report_exc_info, async_report_exc_info, mock_log
    ):
        import rollbar
        from rollbar.contrib.asgi.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        rollbar.SETTINGS['handler'] = 'threading'
        testapp = ReporterMiddleware(FailingTestASGIApp())

        with self.assertRaises(RuntimeError):
            run(testapp({'type': 'http'}, None, None))

        self.assertFalse(async_report_exc_info.called)
        self.assertTrue(sync_report_exc_info.called)
        mock_log.assert_called_once_with(
            'Failed to report asynchronously. Trying to report synchronously.'
        )

    def test_should_support_http_only(self):
        from rollbar.contrib.asgi.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        testapp = ReporterMiddleware(FailingTestASGIApp())

        with mock.patch('rollbar.report_exc_info') as mock_report:
            with self.assertRaises(RuntimeError):
                run(testapp({'type': 'http'}, None, None))

            self.assertTrue(mock_report.called)

        with mock.patch('rollbar.report_exc_info') as mock_report:
            with self.assertRaises(RuntimeError):
                run(testapp({'type': 'websocket'}, None, None))

            self.assertFalse(mock_report.called)

    def test_should_support_type_hints(self):
        from rollbar.contrib.asgi.types import Receive, Scope, Send

        self.assertDictEqual(
            rollbar.contrib.asgi.ReporterMiddleware.__call__.__annotations__,
            {'scope': Scope, 'receive': Receive, 'send': Send, 'return': None},
        )



from collections.abc import Mapping

from rollbar.lib import transforms
from rollbar.lib.transforms.scrub_redact import ScrubRedactTransform, REDACT_REF

from rollbar.test import BaseTest


class NotRedactRef():
    pass

NOT_REDACT_REF = NotRedactRef()

try:
    SCRUBBED = '*' * len(REDACT_REF)
except:
    SCRUBBED = '*' * len(str(REDACT_REF))


class ScrubRedactTransformTest(BaseTest):
    def _assertScrubbed(self, start, expected, redact_char='*', skip_id_check=False):
        scrubber = ScrubRedactTransform(redact_char=redact_char, randomize_len=False)
        result = transforms.transform(start, scrubber)

        if not skip_id_check:
            self.assertNotEqual(id(result), id(expected))

        self.assertEqual(type(result), type(expected))

        if isinstance(result, Mapping):
            self.assertDictEqual(result, expected)
        elif isinstance(result, tuple):
            self.assertTupleEqual(result, expected)
        elif isinstance(result, list):
            self.assertListEqual(result, expected)
        elif isinstance(result, set):
            self.assertSetEqual(result, expected)
        else:
            self.assertEqual(result, expected)

    def test_no_scrub(self):
        obj = NOT_REDACT_REF
        expected = NOT_REDACT_REF
        self._assertScrubbed(obj, expected, skip_id_check=True)

    def test_scrub(self):
        obj = REDACT_REF
        expected = SCRUBBED
        self._assertScrubbed(obj, expected)

    def test_scrub_list(self):
        obj = [REDACT_REF, REDACT_REF, REDACT_REF]
        expected = [SCRUBBED, SCRUBBED, SCRUBBED]
        self._assertScrubbed(obj, expected)

    def test_scrub_set(self):
        obj = set([REDACT_REF, NOT_REDACT_REF])
        expected = set([SCRUBBED, NOT_REDACT_REF])
        self._assertScrubbed(obj, expected)

    def scrub_tuple(self):
        obj = (REDACT_REF, REDACT_REF, REDACT_REF)
        expected = (SCRUBBED, SCRUBBED, SCRUBBED)
        self._assertScrubbed(obj, expected)

    def test_scrub_dict(self):
        obj = {'scrub_me': REDACT_REF}
        expected = {'scrub_me': SCRUBBED}
        self._assertScrubbed(obj, expected)



# NOTE: pyrollbar requires both `Flask` and `blinker` packages to be installed first
from flask import Flask
from flask import got_request_exception

import rollbar
import rollbar.contrib.flask


app = Flask(__name__)


with app.app_context():
    rollbar.init('ACCESS_TOKEN', environment='development')
    # send exceptions from `app` to rollbar, using flask's signal system.
    got_request_exception.connect(rollbar.contrib.flask.report_exception, app)

@app.route('/')
def root():
    foo()
    return '<html><body>Hello World</body></html>'

if __name__ == '__main__':
    app.run()



"""
Tests for Flask instrumentation
"""

import json
import sys
import os

from unittest import mock

import rollbar

from rollbar.test import BaseTest

# access token for https://rollbar.com/rollbar/pyrollbar
TOKEN = '92c10f5616944b81a2e6f3c6493a0ec2'

# Flask doesn't work on python 3.2, so don't test there.
ALLOWED_PYTHON_VERSION = not (sys.version_info[0] == 3 and sys.version_info[1] == 2)

try:
    import flask
    FLASK_INSTALLED = True
except ImportError:
    FLASK_INSTALLED = False


def create_app():
    from flask import Flask, Request, got_request_exception
    import rollbar.contrib.flask
    app = Flask(__name__)

    @app.route('/')
    def index():
        return 'Index page'

    @app.route('/cause_error', methods=['GET', 'POST'])
    def cause_error():
        raise Exception("Uh oh")

    class CustomRequest(Request):
        @property
        def rollbar_person(self):
            return {'id': '123', 'username': 'testuser', 'email': 'test@example.com'}

    app.request_class = CustomRequest

    return app

def init_rollbar(app):
    from flask import got_request_exception
    rollbar._initialized = False
    rollbar.init(TOKEN, 'flasktest',
                 root=os.path.dirname(os.path.realpath(__file__)),
                 allow_logging_basic_config=True,
                 capture_email=True,
                 capture_username=True)
    got_request_exception.connect(rollbar.contrib.flask.report_exception, app)

if ALLOWED_PYTHON_VERSION and FLASK_INSTALLED:
    class FlaskTest(BaseTest):
        def setUp(self):
            super(FlaskTest, self).setUp()
            self.app = create_app()
            init_rollbar(self.app)
            self.client = self.app.test_client()

        def test_index(self):
            resp = self.client.get('/')
            self.assertEqual(resp.status_code, 200)

        def assertStringEqual(self, left, right):
            if sys.version_info[0] > 2:
                if hasattr(left, 'decode'):
                    left = left.decode('ascii')
                if hasattr(right, 'decode'):
                    right = right.decode('ascii')

                return self.assertEqual(left, right)
            else:
                return self.assertEqual(left, right)

        @mock.patch('rollbar.send_payload')
        def test_uncaught(self, send_payload):
            resp = self.client.get('/cause_error?foo=bar',
                headers={'X-Real-Ip': '1.2.3.4', 'User-Agent': 'Flask Test'})
            self.assertEqual(resp.status_code, 500)

            self.assertEqual(send_payload.called, True)
            payload = send_payload.call_args[0][0]
            data = payload['data']

            self.assertIn('body', data)
            self.assertEqual(data['body']['trace']['exception']['class'], 'Exception')
            self.assertStringEqual(data['body']['trace']['exception']['message'], 'Uh oh')

            self.assertIn('person', data)
            self.assertDictEqual(data['person'],
                {'id': '123', 'username': 'testuser', 'email': 'test@example.com'})

            self.assertIn('request', data)
            self.assertEqual(data['request']['url'], 'http://localhost/cause_error?foo=bar')

            # The behavior of implicitly converting werkzeug.ImmutableMultiDict
            # using dict() changes starting in Python 3.6.
            # See _build_werkzeug_request_data()
            if sys.version_info >= (3, 6):
                self.assertDictEqual(data['request']['GET'], {'foo': 'bar'})
            else:
                self.assertDictEqual(data['request']['GET'], {'foo': ['bar']})

            self.assertEqual(data['request']['user_ip'], '1.2.3.4')
            self.assertEqual(data['request']['method'], 'GET')
            self.assertEqual(data['request']['headers']['User-Agent'], 'Flask Test')

        @mock.patch('rollbar.send_payload')
        def test_uncaught_json_request(self, send_payload):
            json_body = {"hello": "world"}
            json_body_str = json.dumps(json_body)
            resp = self.client.post('/cause_error', data=json_body_str,
                headers={'Content-Type': 'application/json', 'X-Forwarded-For': '5.6.7.8'})

            self.assertEqual(resp.status_code, 500)

            self.assertEqual(send_payload.called, True)
            payload = send_payload.call_args[0][0]
            data = payload['data']

            self.assertIn('body', data)
            self.assertEqual(data['body']['trace']['exception']['class'], 'Exception')
            self.assertStringEqual(data['body']['trace']['exception']['message'], 'Uh oh')

            self.assertIn('person', data)
            self.assertDictEqual(data['person'],
                {'id': '123', 'username': 'testuser', 'email': 'test@example.com'})

            self.assertIn('request', data)
            self.assertEqual(data['request']['url'], 'http://localhost/cause_error')
            self.assertEqual(data['request']['body'], json_body)
            self.assertEqual(data['request']['user_ip'], '5.6.7.8')
            self.assertEqual(data['request']['method'], 'POST')

        @mock.patch('rollbar.send_payload')
        def test_uncaught_no_username_no_email(self, send_payload):
            rollbar.SETTINGS['capture_email'] = False
            rollbar.SETTINGS['capture_username'] = False

            resp = self.client.get('/cause_error?foo=bar',
                headers={'X-Real-Ip': '1.2.3.4', 'User-Agent': 'Flask Test'})
            self.assertEqual(resp.status_code, 500)

            self.assertEqual(send_payload.called, True)
            payload = send_payload.call_args[0][0]
            data = payload['data']

            self.assertIn('body', data)
            self.assertEqual(data['body']['trace']['exception']['class'], 'Exception')
            self.assertStringEqual(data['body']['trace']['exception']['message'], 'Uh oh')

            self.assertIn('person', data)
            self.assertDictEqual(data['person'],
                {'id': '123', 'username': None, 'email': None})

            self.assertIn('request', data)
            self.assertEqual(data['request']['url'], 'http://localhost/cause_error?foo=bar')

            # The behavior of implicitly converting werkzeug.ImmutableMultiDict
            # using dict() changes starting in Python 3.6.
            # See _build_werkzeug_request_data()
            if sys.version_info >= (3, 6):
                self.assertDictEqual(data['request']['GET'], {'foo': 'bar'})
            else:
                self.assertDictEqual(data['request']['GET'], {'foo': ['bar']})

            self.assertEqual(data['request']['user_ip'], '1.2.3.4')
            self.assertEqual(data['request']['method'], 'GET')
            self.assertEqual(data['request']['headers']['User-Agent'], 'Flask Test')

            rollbar.SETTINGS['capture_email'] = True
            rollbar.SETTINGS['capture_username'] = True



import random

from rollbar.lib import build_key_matcher
from rollbar.lib.transform import Transform


class ScrubTransform(Transform):
    suffix_matcher = None
    def __init__(self, suffixes=None, redact_char='*', randomize_len=True):
        super(ScrubTransform, self).__init__()
        if suffixes is not None and len(suffixes) > 0:
            self.suffix_matcher = build_key_matcher(suffixes, type='suffix')
        self.redact_char = redact_char
        self.randomize_len = randomize_len

    def in_scrub_fields(self, key):
        if self.suffix_matcher is None:
            return False
        return self.suffix_matcher(key)

    def redact(self, val):
        if self.randomize_len:
            _len = random.randint(3, 20)
        else:
            try:
                _len = len(val)
            except:
                _len = len(str(val))

        return self.redact_char * _len

    def default(self, o, key=None):
        if self.in_scrub_fields(key):
            return self.redact(o)

        return o


__all__ = ['ScrubTransform']



import copy

from collections.abc import Mapping

from rollbar.lib import transforms
from rollbar.lib.transforms.scrub import ScrubTransform

from rollbar.test import BaseTest, SNOWMAN


class ScrubTransformTest(BaseTest):
    def _assertScrubbed(self, suffixes, start, expected, redact_char='*', skip_id_check=False):
        scrubber = ScrubTransform(suffixes=suffixes, redact_char=redact_char, randomize_len=False)
        result = transforms.transform(start, scrubber)

        """
        print start
        print result
        print expected
        """

        if not skip_id_check:
            self.assertNotEqual(id(result), id(expected))

        self.assertEqual(type(result), type(expected))

        if isinstance(result, Mapping):
            self.assertDictEqual(result, expected)
        elif isinstance(result, tuple):
            self.assertTupleEqual(result, expected)
        elif isinstance(result, (list, set)):
            self.assertListEqual(result, expected)
        else:
            self.assertEqual(result, expected)

    def test_string(self):
        obj = 'simple_string'

        # The Python interpreter is pretty smart... let's fool it into
        # not using the same string object for the expected var
        expected = ('simple_string' + ' dummy').split()[0]
        self._assertScrubbed([['password']], obj, expected)

    def test_num(self):
        obj = 22
        expected = 22
        self._assertScrubbed([['password']], obj, expected, skip_id_check=True)

    def test_None(self):
        obj = None
        expected = None
        self._assertScrubbed([['password']], obj, expected, skip_id_check=True)

    def test_list_of_None(self):
        obj = [None, None, None]
        expected = [None, None, None]
        self._assertScrubbed([['password']], obj, expected)

    def test_no_suffixes(self):
        obj = {'hello': 'world', 'password': 'cleartext'}
        expected = dict(obj)
        self._assertScrubbed([], obj, expected)

    def test_scrub_simple_dict(self):
        obj = {'hello': 'world', 'password': 'cleartext'}
        expected = copy.deepcopy(obj)
        expected['password'] = '*********'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_dict_in_list(self):
        obj = [{'hello': 'world', 'password': 'cleartext'}, 'another element']
        expected = copy.deepcopy(obj)
        expected[0]['password'] = '*********'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_dict_in_tuple(self):
        obj = ({'hello': 'world', 'password': 'cleartext'}, 'another element')
        expected = copy.deepcopy(obj)
        expected[0]['password'] = '*********'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_dict_in_dict(self):
        obj = {'the data': {'hello': 'world', 'password': 'cleartext'}, 'some other data': 3}
        expected = copy.deepcopy(obj)
        expected['the data']['password'] = '*********'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_multiple_suffixes(self):
        obj = {'hello': 'world', 'password': 'cleartext'}
        expected = copy.deepcopy(obj)
        expected['hello'] = '*****'
        expected['password'] = '*********'
        self._assertScrubbed([['hello'], ['password']], obj, expected)

    def test_scrub_no_suffix_match(self):
        obj = {'hello': 'world', 'password': 'cleartext'}
        expected = copy.deepcopy(obj)
        self._assertScrubbed([['not found'], ['access_token']], obj, expected)

    def test_scrub_full_dict(self):
        obj = {'hello': 'world', 'password': {'nested': 'secrets'}}
        expected = copy.deepcopy(obj)
        expected['password'] = '*'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_full_list(self):
        obj = {'hello': 'world', 'password': [1, 2, 3, 4]}
        expected = copy.deepcopy(obj)
        expected['password'] = '****'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_full_dict_with_nested_secret(self):
        obj = {'hello': 'world', 'password': [{'password': 'secret', 'nested': {'password': 'secret!!!'}}]}
        expected = copy.deepcopy(obj)
        expected['password'] = '*'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_one_secret_and_one_clear_reference(self):
        ref = {'scrub': 'me', 'some': 'times'}
        obj = {'hello': 'world', 'password': ref, 'should be clear': ref}
        expected = copy.deepcopy(obj)
        expected['password'] = '**'
        self._assertScrubbed([['password']], obj, expected)

    def test_scrub_circular(self):
        ref = {'scrub': 'me', 'some': 'times'}
        obj = {'hello': 'world', 'password': ref}
        ref['circular'] = obj
        """
        Obj A =
        {
          'hello': 'world',
          'password': {
            'scrub': 'me', 'some': 'times', 'circular': A
          }
        }
        """
        expected = copy.deepcopy(obj)
        expected['password'] = '***'

        self._assertScrubbed([['password']], obj, expected)

    def test_circular(self):
        ref = {'scrub': 'me', 'some': 'times'}
        obj = {'hello': 'world', 'password': ref}
        ref['circular'] = obj

        scrubber = ScrubTransform([])
        result = transforms.transform(obj, scrubber)

        self.assertIsNot(result, obj)
        self.assertIsNot(result['password'], ref)
        self.assertIsNot(result['password']['circular'], obj)
        self.assertIs(result['password']['circular']['password'], result['password']['circular']['password']['circular']['password'])

    def test_unicode_keys(self):
        obj = {
            SNOWMAN: 'hello'
        }
        expected = copy.deepcopy(obj)
        self._assertScrubbed([], obj, expected)

    def test_scrub_unicode_keys(self):
        obj = {
            SNOWMAN: 'hello'
        }
        expected = {
            SNOWMAN: '*****'
        }
        self._assertScrubbed([[SNOWMAN]], obj, expected)

    def test_scrub_unicode_values(self):
        obj = {
            'password': SNOWMAN
        }
        expected = {
            'password': '***'
        }
        self._assertScrubbed([['password']], obj, expected)



from urllib.parse import urlparse, parse_qs

from rollbar.lib import transforms, string_types
from rollbar.lib.transforms.scruburl import ScrubUrlTransform, _starts_with_auth_re

from rollbar.test import BaseTest, SNOWMAN_UNICODE


SNOWMAN_LEN = len(SNOWMAN_UNICODE)


class ScrubUrlTransformTest(BaseTest):
    def _assertScrubbed(self,
                        params_to_scrub,
                        start,
                        expected,
                        scrub_username=False,
                        scrub_password=True,
                        redact_char='-',
                        skip_id_check=False):
        scrubber = ScrubUrlTransform(suffixes=[],
                                     params_to_scrub=params_to_scrub,
                                     scrub_username=scrub_username,
                                     scrub_password=scrub_password,
                                     redact_char=redact_char,
                                     randomize_len=False)
        result = transforms.transform(start, scrubber)

        if not skip_id_check:
            self.assertNotEqual(id(result), id(expected))

        self.assertEqual(type(expected), type(result))
        self.assertIsInstance(result, string_types)
        self._compare_urls(expected, result)

    def _compare_urls(self, url1, url2):
        if _starts_with_auth_re.match(url1):
            url1 = '//%s' % url1

        if _starts_with_auth_re.match(url2):
            url2 = '//%s' % url2

        parsed_urls = [urlparse(url) for url in (url1, url2)]
        qs_params = [parse_qs(x.query, keep_blank_values=True) for x in parsed_urls]
        num_params = [len(x) for x in qs_params]
        param_names = [set(x.keys()) for x in qs_params]

        self.assertEqual(*num_params)
        self.assertDictEqual(*qs_params)
        self.assertSetEqual(*param_names)

        for facet in ('scheme', 'netloc', 'path', 'params', 'username', 'password', 'hostname', 'port'):
            comp = [getattr(x, facet) for x in parsed_urls]
            self.assertEqual(*comp)

    def test_no_scrub(self):
        obj = 'http://hello.com/?foo=bar'
        expected = obj
        self._assertScrubbed(['password'], obj, expected, skip_id_check=True)

    def test_not_url(self):
        obj = 'I am a plain\'ol string'
        expected = obj
        self._assertScrubbed(['password'], obj, expected, skip_id_check=True)

    def test_scrub_simple_url_params(self):
        obj = 'http://foo.com/asdf?password=secret'
        expected = obj.replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected)

    def test_scrub_utf8_url_params(self):
        obj = 'http://foo.com/asdf?password=%s' % SNOWMAN_UNICODE
        expected = obj.replace(SNOWMAN_UNICODE, '-' * SNOWMAN_LEN)
        self._assertScrubbed(['password'], obj, expected)

    def test_scrub_utf8_url_keys(self):
        obj = 'http://foo.com/asdf?%s=secret' % SNOWMAN_UNICODE
        expected = obj.replace('secret', '------')
        self._assertScrubbed([str(SNOWMAN_UNICODE)], obj, expected)

    def test_scrub_multi_url_params(self):
        obj = 'http://foo.com/asdf?password=secret&password=secret2&token=TOK&clear=text'
        expected = obj.replace('secret2', '-------').replace('secret', '------').replace('TOK', '---')
        self._assertScrubbed(['password', 'token'], obj, expected)

    def test_scrub_password_auth(self):
        obj = 'http://cory:secr3t@foo.com/asdf?password=secret&clear=text'
        expected = obj.replace('secr3t', '------').replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected)

    def test_scrub_username_auth(self):
        obj = 'http://cory:secr3t@foo.com/asdf?password=secret&clear=text'
        expected = obj.replace('cory', '----').replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected, scrub_password=False, scrub_username=True)

    def test_scrub_username_and_password_auth(self):
        obj = 'http://cory:secr3t@foo.com/asdf?password=secret&clear=text'
        expected = obj.replace('cory', '----').replace('secr3t', '------').replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected, scrub_password=True, scrub_username=True)

    def test_scrub_missing_scheme(self):
        obj = '//cory:secr3t@foo.com/asdf?password=secret&clear=text'
        expected = obj.replace('secr3t', '------').replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected)

    def test_scrub_missing_scheme_and_double_slash(self):
        obj = 'cory:secr3t@foo.com/asdf?password=secret&clear=text'
        expected = obj.replace('secr3t', '------').replace('secret', '------')
        self._assertScrubbed(['password'], obj, expected)

    def test_keep_blank_url_params(self):
        obj = 'http://foo.com/asdf?foo=bar&baz='
        expected = obj
        self._assertScrubbed(['password'], obj, expected, skip_id_check=True)

    def test_scrub_dict_val_isnt_string(self):

        url = 'cory:secr3t@foo.com/asdf?password=secret&clear=text'

        # Every string which is a URL should be scrubbed
        obj = {
            'url': [url]
        }

        scrubber = ScrubUrlTransform(suffixes=[('url',)], params_to_scrub=['password'], randomize_len=False)
        result = transforms.transform(obj, scrubber)

        expected = url.replace('secr3t', '------').replace('secret', '------')
        self._assertScrubbed(['password'], result['url'][0], expected)

    def test_scrub_dict_nested_key_match_with_circular_ref(self):
        # If a URL is a circular reference then let's make sure to
        # show the scrubbed, original URL
        url = 'cory:secr3t@foo.com/asdf?password=secret&clear=text'
        obj = {
            'url': [{'link': url}],
            'link': [{'url': url}]
        }

        scrubber = ScrubUrlTransform(suffixes=[('url',), ('link',)], params_to_scrub=['password'], randomize_len=False)
        result = transforms.transform(obj, scrubber)

        self.assertNotIn('secr3t', result['url'][0]['link'])
        self.assertNotIn('secret', result['url'][0]['link'])
        self.assertNotIn('secr3t', result['link'][0]['url'])
        self.assertNotIn('secret', result['link'][0]['url'])
        self.assertNotRegex(result['url'][0]['link'], r'^-+$')
        self.assertNotRegex(result['link'][0]['url'], r'^-+$')



import sys

try:
    import starlette

    STARLETTE_INSTALLED = True
except ImportError:
    STARLETTE_INSTALLED = False

import unittest

from rollbar.test import BaseTest
from rollbar.test.utils import get_public_attrs

ALLOWED_PYTHON_VERSION = sys.version_info >= (3, 6)


@unittest.skipUnless(
    STARLETTE_INSTALLED and ALLOWED_PYTHON_VERSION,
    'Global request access requires Python3.6+',
)
class RequestTest(BaseTest):
    def test_should_accept_request_param(self):
        from starlette.requests import Request
        from rollbar.contrib.starlette.requests import store_current_request
        from rollbar.lib._async import async_receive

        scope = {
            'client': ['testclient', 50000],
            'headers': [
                (b'host', b'testserver'),
                (b'user-agent', b'testclient'),
                (b'accept-encoding', b'gzip, deflate'),
                (b'accept', b'*/*'),
                (b'connection', b'keep-alive'),
            ],
            'http_version': '1.1',
            'method': 'GET',
            'path': '/',
            'query_string': b'',
            'root_path': '',
            'scheme': 'http',
            'server': ['testserver', 80],
            'type': 'http',
        }
        receive = async_receive(
            {'type': 'http.request', 'body': b'body body', 'mode_body': False}
        )
        request = Request(scope, receive)

        stored_request = store_current_request(request)

        self.assertEqual(get_public_attrs(request), get_public_attrs(stored_request))

    def test_should_accept_scope_param_if_http_type(self):
        from starlette.requests import Request
        from rollbar.contrib.starlette.requests import store_current_request
        from rollbar.lib._async import async_receive

        scope = {
            'client': ['testclient', 50000],
            'headers': [
                (b'host', b'testserver'),
                (b'user-agent', b'testclient'),
                (b'accept-encoding', b'gzip, deflate'),
                (b'accept', b'*/*'),
                (b'connection', b'keep-alive'),
            ],
            'http_version': '1.1',
            'method': 'GET',
            'path': '/',
            'query_string': b'',
            'root_path': '',
            'scheme': 'http',
            'server': ['testserver', 80],
            'type': 'http',
        }
        receive = async_receive(
            {'type': 'http.request', 'body': b'body body', 'mode_body': False}
        )
        expected_request = Request(scope, receive)

        request = store_current_request(scope, receive)

        self.assertEqual(get_public_attrs(request), get_public_attrs(expected_request))

    def test_should_not_accept_scope_param_if_not_http_type(self):
        from rollbar.contrib.starlette.requests import store_current_request

        scope = {'asgi': {'spec_version': '2.0', 'version': '3.0'}, 'type': 'lifespan'}
        receive = {}

        request = store_current_request(scope, receive)

        self.assertIsNone(request)

    def test_hasuser(self):
        from starlette.requests import Request
        from rollbar.contrib.starlette.requests import hasuser

        request = Request({'type': 'http'}, {})
        self.assertFalse(hasuser(request))

        request = Request({'type': 'http', 'user': 'testuser'}, {})
        self.assertTrue(hasuser(request))
        self.assertEqual(request.user, 'testuser')



__all__ = ['get_current_request']

import logging
import sys
from typing import Optional, Union

from starlette.requests import Request
from starlette.types import Receive, Scope

log = logging.getLogger(__name__)

if sys.version_info[:2] == (3, 6):
    # Backport PEP 567
    try:
        import aiocontextvars
    except ImportError:
        # Do not raise an exception as the module is exported to package API
        # but is still optional
        log.error(
            'Python 3.6 requires `aiocontextvars` package to be installed'
            ' to support global access to request objects'
        )

try:
    from contextvars import ContextVar
except ImportError:
    ContextVar = None

if ContextVar:
    _current_request: ContextVar[Optional[Request]] = ContextVar(
        'rollbar-request-object', default=None
    )


def get_current_request() -> Optional[Request]:
    """
    Return current request.

    Do NOT modify the returned request object.
    """

    if ContextVar is None:
        log.error(
            'Python 3.7+ (or aiocontextvars package)'
            ' is required to receive current request.'
        )
        return None

    return _current_request.get()


def store_current_request(
    request_or_scope: Union[Request, Scope], receive: Optional[Receive] = None
) -> Optional[Request]:
    if ContextVar is None:
        return None

    if receive is None:
        request = request_or_scope
    elif request_or_scope['type'] == 'http':
        request = Request(request_or_scope, receive)
    else:
        request = None

    _current_request.set(request)
    return request


def hasuser(request: Request) -> bool:
    try:
        return hasattr(request, 'user')
    except AssertionError:
        return False



__all__ = ['ReporterMiddleware']

from fastapi import __version__

from rollbar.contrib.asgi.integration import integrate
from rollbar.contrib.starlette import ReporterMiddleware as StarletteReporterMiddleware


@integrate(framework_name=f'fastapi {__version__}')
class ReporterMiddleware(StarletteReporterMiddleware):
    ...



import asyncio
import contextlib
import inspect
import logging
import sys
from unittest import mock
from urllib.parse import urljoin

try:
    import httpx
except ImportError:
    httpx = None

import rollbar
from rollbar import DEFAULT_TIMEOUT
from rollbar.lib import transport

log = logging.getLogger(__name__)

ALLOWED_HANDLERS = (
    'async',
    'httpx',
)


if sys.version_info[:2] == (3, 6):
    # Backport PEP 567
    try:
        import aiocontextvars
    except ImportError:
        log.warning(
            'Python3.6 does not provide the `contextvars` module.'
            ' Some advanced features may not work as expected.'
            ' Please upgrade Python or install `aiocontextvars`.'
        )

try:
    from contextvars import ContextVar
except ImportError:
    ContextVar = None

if ContextVar:
    _ctx_handler = ContextVar('rollbar-handler', default=None)
else:
    _ctx_handler = None


class RollbarAsyncError(Exception):
    ...


async def report_exc_info(
    exc_info=None, request=None, extra_data=None, payload_data=None, level=None, **kw
):
    """
    Asynchronously reports an exception to Rollbar, using exc_info (from calling sys.exc_info())

    exc_info: optional, should be the result of calling sys.exc_info(). If omitted, sys.exc_info() will be called here.
    request: optional, a Starlette, WebOb, Werkzeug-based or Sanic request object.
    extra_data: optional, will be included in the 'custom' section of the payload
    payload_data: optional, dict that will override values in the final payload
                  (e.g. 'level' or 'fingerprint')
    kw: provided for legacy purposes; unused.

    Example usage:

    rollbar.init(access_token='YOUR_PROJECT_ACCESS_TOKEN')

    async def func():
        try:
            do_something()
        except:
            await report_exc_info(sys.exc_info(), request, {'foo': 'bar'}, {'level': 'warning'})
    """
    with AsyncHandler():
        try:
            return await call_later(
                _report_exc_info(
                    exc_info, request, extra_data, payload_data, level, **kw
                )
            )
        except Exception as e:
            log.exception('Exception while reporting exc_info to Rollbar. %r', e)


async def report_message(
    message, level='error', request=None, extra_data=None, payload_data=None, **kw
):
    """
    Asynchronously reports an arbitrary string message to Rollbar.

    message: the string body of the message
    level: level to report at. One of: 'critical', 'error', 'warning', 'info', 'debug'
    request: the request object for the context of the message
    extra_data: dictionary of params to include with the message. 'body' is reserved.
    payload_data: param names to pass in the 'data' level of the payload; overrides defaults.
    """

    with AsyncHandler():
        try:
            return await call_later(
                _report_message(message, level, request, extra_data, payload_data)
            )
        except Exception as e:
            log.exception('Exception while reporting message to Rollbar. %r', e)


async def _report_exc_info(
    exc_info=None, request=None, extra_data=None, payload_data=None, level=None, **kw
):
    return rollbar.report_exc_info(
        exc_info, request, extra_data, payload_data, level, **kw
    )


async def _report_message(
    message, level='error', request=None, extra_data=None, payload_data=None, **kw
):
    return rollbar.report_message(message, level, request, extra_data, payload_data)


async def _post_api_httpx(path, payload_str, access_token=None):
    headers = {'Content-Type': 'application/json'}
    if access_token is not None:
        headers['X-Rollbar-Access-Token'] = access_token
    else:
        headers['X-Rollbar-Access-Token'] = rollbar.SETTINGS.get('access_token')

    proxy_cfg = {
        'proxy': rollbar.SETTINGS.get('http_proxy'),
        'proxy_user': rollbar.SETTINGS.get('http_proxy_user'),
        'proxy_password': rollbar.SETTINGS.get('http_proxy_password'),
    }
    proxies = transport._get_proxy_cfg(proxy_cfg)

    url = urljoin(rollbar.SETTINGS['endpoint'], path)
    async with httpx.AsyncClient(
        proxies=proxies, verify=rollbar.SETTINGS.get('verify_https', True)
    ) as client:
        resp = await client.post(
            url,
            content=payload_str,
            headers=headers,
            timeout=rollbar.SETTINGS.get('timeout', DEFAULT_TIMEOUT),
        )

    try:
        return rollbar._parse_response(path, access_token, payload_str, resp)
    except Exception as e:
        log.exception('Exception while posting item %r', e)


async def try_report(
    exc_info=None, request=None, extra_data=None, payload_data=None, level=None, **kw
):
    current_handler = rollbar.SETTINGS.get('handler')
    if not (current_handler in ALLOWED_HANDLERS or current_handler == 'default'):
        raise RollbarAsyncError('No async handler set.')

    if httpx is None:
        raise RollbarAsyncError('HTTPX is required')

    return await report_exc_info(
        exc_info, request, extra_data, payload_data, level, **kw
    )


class AsyncHandler:
    def __init__(self):
        self.global_handler = None
        self.token = None

    def with_ctx_handler(self):
        if self.global_handler in ALLOWED_HANDLERS:
            self.token = _ctx_handler.set(self.global_handler)
        else:
            log.warning(
                'Running coroutines requires async compatible handler. Switching to default async handler.'
            )
            self.token = _ctx_handler.set('async')

        return _ctx_handler.get()

    def with_global_handler(self):
        return self.global_handler

    def __enter__(self):
        self.global_handler = rollbar.SETTINGS.get('handler')

        if _ctx_handler:
            return self.with_ctx_handler()
        else:
            return self.with_global_handler()

    def __exit__(self, exc_type, exc_value, traceback):
        if _ctx_handler and self.token:
            _ctx_handler.reset(self.token)


def get_current_handler():
    if _ctx_handler is None:
        return rollbar.SETTINGS.get('handler')

    handler = _ctx_handler.get()

    if handler is None:
        return rollbar.SETTINGS.get('handler')

    return handler


def call_later(coro):
    if sys.version_info < (3, 7):
        return asyncio.ensure_future(coro)

    return asyncio.create_task(coro)


# test helpers
# TODO: move to rollbar.test.async_helper after migrating from unittest


def run(coro):
    if sys.version_info >= (3, 7):
        return asyncio.run(coro)

    assert inspect.iscoroutine(coro)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coro)
    finally:
        loop.close()
        asyncio.set_event_loop(None)


def async_receive(message):
    async def receive():
        return message

    assert message['type'] == 'http.request'
    return receive


async def coroutine():
    ...


class BareMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        await self.app(scope, receive, send)


class FailingTestASGIApp:
    async def __call__(self, scope, receive, send):
        await self.app(scope, receive, send)

    async def app(self, scope, receive, send):
        raise RuntimeError('Invoked only for testing')


# for Python 3.7- compatibility
class AsyncMock(mock.MagicMock):
    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)



#!/usr/bin/env python

# This example uses Uvicorn package that must be installed. However, it can be
# replaced with any other ASGI-compliant server.
#
# Optional asynchronous reporting requires HTTPX package to be installed.
#
# NOTE: Starlette middlewares don't allow to collect streamed content like a request body.
#
# Run: python app.py

import rollbar
import uvicorn

from rollbar.contrib.starlette import ReporterMiddleware as RollbarMiddleware
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse

# Initialize Rollbar SDK with your server-side ACCESS_TOKEN
rollbar.init(
    'ACCESS_TOKEN',
    environment='staging',
    handler='async',  # For asynchronous reporting use: default, async or httpx
)

# Integrate Rollbar with Starlette application
app = Starlette()
app.add_middleware(RollbarMiddleware)  # should be added as the first middleware


# Verify application runs correctly
#
# $ curl http://localhost:8888
@app.route('/')
async def root(request):
    return PlainTextResponse('hello world')


# Cause an uncaught exception to be sent to Rollbar
# GET query params will be sent to Rollbar and available in the UI
#
# $ curl http://localhost:8888/error?param1=hello&param2=world
async def localfunc(arg1, arg2, arg3):
    # Both local variables and function arguments will be sent to Rollbar
    # and available in the UI
    localvar = 'local variable'
    cause_error_with_local_variables


@app.route('/error')
async def error(request):
    await localfunc('func_arg1', 'func_arg2', 1)
    return PlainTextResponse("You shouldn't be seeing this")


if __name__ == '__main__':
    uvicorn.run(app, host='localhost', port=8888)



import copy
import importlib
import sys

from unittest import mock

try:
    import starlette

    STARLETTE_INSTALLED = True
except ImportError:
    STARLETTE_INSTALLED = False

import unittest

import rollbar
from rollbar.lib._async import AsyncMock
from rollbar.test import BaseTest
from rollbar.test.utils import get_public_attrs

ALLOWED_PYTHON_VERSION = sys.version_info >= (3, 6)


@unittest.skipUnless(
    STARLETTE_INSTALLED and ALLOWED_PYTHON_VERSION, 'Starlette requires Python3.6+'
)
class ReporterMiddlewareTest(BaseTest):
    default_settings = copy.deepcopy(rollbar.SETTINGS)

    def setUp(self):
        importlib.reload(rollbar)
        rollbar.SETTINGS = copy.deepcopy(self.default_settings)
        rollbar.SETTINGS['handler'] = 'async'

    @mock.patch('rollbar.report_exc_info')
    def test_should_catch_and_report_errors(self, mock_report):
        from starlette.applications import Starlette
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/')

        mock_report.assert_called_once()

        args, kwargs = mock_report.call_args
        self.assertEqual(kwargs, {})

        exc_type, exc_value, exc_tb = args[0]

        self.assertEqual(exc_type, ZeroDivisionError)
        self.assertIsInstance(exc_value, ZeroDivisionError)

    @mock.patch('rollbar.report_exc_info')
    def test_should_report_with_request_data(self, mock_report):
        from starlette.applications import Starlette
        from starlette.requests import Request
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/')

        mock_report.assert_called_once()
        request = mock_report.call_args[0][1]

        self.assertIsInstance(request, Request)

    @mock.patch('rollbar._check_config', return_value=True)
    @mock.patch('rollbar._serialize_frame_data')
    @mock.patch('rollbar.send_payload')
    def test_should_send_payload_with_request_data(self, mock_send_payload, *mocks):
        from starlette.applications import Starlette
        from starlette.requests import Request
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/{path}')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/test?param1=value1&param2=value2')

        mock_send_payload.assert_called_once()
        payload = mock_send_payload.call_args[0][0]
        payload_request = payload['data']['request']

        self.assertEqual(payload_request['method'], 'GET')
        self.assertEqual(payload_request['user_ip'], 'testclient')
        self.assertEqual(
            payload_request['url'],
            'http://testserver/test?param1=value1&param2=value2',
        )
        self.assertDictEqual(payload_request['params'], {'path': 'test'})
        self.assertDictEqual(
            payload_request['GET'], {'param1': 'value1', 'param2': 'value2'}
        )
        self.assertDictEqual(
            payload_request['headers'],
            {
                'accept': '*/*',
                'accept-encoding': 'gzip, deflate',
                'connection': 'keep-alive',
                'host': 'testserver',
                'user-agent': 'testclient',
            },
        )

    @mock.patch('rollbar._check_config', return_value=True)
    @mock.patch('rollbar.send_payload')
    def test_should_add_framework_version_to_payload(self, mock_send_payload, *mocks):
        import starlette
        from starlette.applications import Starlette
        import rollbar
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        self.assertIsNone(rollbar.BASE_DATA_HOOK)

        app = Starlette()
        app.add_middleware(ReporterMiddleware)
        app.build_middleware_stack()

        rollbar.report_exc_info()

        mock_send_payload.assert_called_once()
        payload = mock_send_payload.call_args[0][0]

        self.assertIn('starlette', payload['data']['framework'])
        self.assertIn(starlette.__version__, payload['data']['framework'])

    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_async_report_exc_info_if_default_handler(
        self, sync_report_exc_info, async_report_exc_info
    ):
        from starlette.applications import Starlette
        from starlette.testclient import TestClient
        import rollbar
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        rollbar.SETTINGS['handler'] = 'default'

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/')

        async_report_exc_info.assert_called_once()
        sync_report_exc_info.assert_not_called()

    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_async_report_exc_info_if_any_async_handler(
        self, sync_report_exc_info, async_report_exc_info
    ):
        from starlette.applications import Starlette
        from starlette.testclient import TestClient
        import rollbar
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        rollbar.SETTINGS['handler'] = 'httpx'

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/')

        async_report_exc_info.assert_called_once()
        sync_report_exc_info.assert_not_called()

    @mock.patch('logging.Logger.warning')
    @mock.patch('rollbar.lib._async.report_exc_info', new_callable=AsyncMock)
    @mock.patch('rollbar.report_exc_info')
    def test_should_use_sync_report_exc_info_if_non_async_handlers(
        self, sync_report_exc_info, async_report_exc_info, mock_log
    ):
        from starlette.applications import Starlette
        from starlette.testclient import TestClient
        import rollbar
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        rollbar.SETTINGS['handler'] = 'threading'

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(request):
            1 / 0

        client = TestClient(app)
        with self.assertRaises(ZeroDivisionError):
            client.get('/')

        sync_report_exc_info.assert_called_once()
        async_report_exc_info.assert_not_called()
        mock_log.assert_called_once_with(
            'Failed to report asynchronously. Trying to report synchronously.'
        )

    @unittest.skipUnless(
        sys.version_info >= (3, 6), 'Global request access requires Python 3.6+'
    )
    @mock.patch('rollbar.contrib.starlette.middleware.store_current_request')
    def test_should_store_current_request(self, store_current_request):
        from starlette.applications import Starlette
        from starlette.responses import PlainTextResponse
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware

        expected_scope = {
            'client': ['testclient', 50000],
            'headers': [
                (b'host', b'testserver'),
                (b'accept', b'*/*'),
                (b'accept-encoding', b'gzip, deflate'),
                (b'connection', b'keep-alive'),
                (b'user-agent', b'testclient'),
            ],
            'http_version': '1.1',
            'method': 'GET',
            'path': '/',
            'query_string': b'',
            'root_path': '',
            'scheme': 'http',
            'server': ['testserver', 80],
            'type': 'http',
        }

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/{param}')
        async def root(request):
            return PlainTextResponse('OK')

        client = TestClient(app)
        client.get('/')

        store_current_request.assert_called_once()

        scope = store_current_request.call_args[0][0]
        self.assertEqual(scope, {**expected_scope, **scope})

    @unittest.skipUnless(
        sys.version_info >= (3, 6), 'Global request access is supported in Python 3.6+'
    )
    def test_should_return_current_request(self):
        from starlette.applications import Starlette
        from starlette.responses import PlainTextResponse
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware
        from rollbar.contrib.starlette import get_current_request

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(original_request):
            request = get_current_request()

            self.assertEqual(get_public_attrs(request), get_public_attrs(original_request))

            return PlainTextResponse('OK')

        client = TestClient(app)
        client.get('/')

    @mock.patch('rollbar.contrib.starlette.requests.ContextVar', None)
    @mock.patch('logging.Logger.error')
    def test_should_not_return_current_request_for_older_python(self, mock_log):
        from starlette.applications import Starlette
        from starlette.responses import PlainTextResponse
        from starlette.testclient import TestClient
        from rollbar.contrib.starlette.middleware import ReporterMiddleware
        from rollbar.contrib.starlette import get_current_request

        app = Starlette()
        app.add_middleware(ReporterMiddleware)

        @app.route('/')
        async def root(original_request):
            request = get_current_request()

            self.assertIsNone(request)
            self.assertNotEqual(request, original_request)
            mock_log.assert_called_once_with(
                'Python 3.7+ (or aiocontextvars package)'
                ' is required to receive current request.'
            )

            return PlainTextResponse('OK')

        client = TestClient(app)
        client.get('/')

    def test_should_support_http_only(self):
        from rollbar.contrib.starlette.middleware import ReporterMiddleware
        from rollbar.lib._async import FailingTestASGIApp, run

        testapp = ReporterMiddleware(FailingTestASGIApp())

        with mock.patch('rollbar.report_exc_info') as mock_report:
            with self.assertRaises(RuntimeError):
                run(testapp({'type': 'http'}, None, None))

            mock_report.assert_called_once()

        with mock.patch('rollbar.report_exc_info') as mock_report:
            with self.assertRaises(RuntimeError):
                run(testapp({'type': 'websocket'}, None, None))

            mock_report.assert_not_called()

    def test_should_support_type_hints(self):
        from starlette.types import Receive, Scope, Send

        self.assertDictEqual(
            rollbar.contrib.starlette.ReporterMiddleware.__call__.__annotations__,
            {'scope': Scope, 'receive': Receive, 'send': Send, 'return': None},
        )



from rollbar.lib import events, filters

from rollbar.test import BaseTest


class BasicFiltersTest(BaseTest):
    def setUp(self):
        events.reset()
        filters.add_builtin_filters({})

    def test_rollbar_ignored_exception(self):
        class IgnoredException(Exception):
            _rollbar_ignore = True

        class NotIgnoredException(Exception):
            _rollbar_ignore = False

        self.assertFalse(events.on_exception_info((None, IgnoredException(), None)))
        self.assertIsNot(events.on_exception_info((None, NotIgnoredException(), None)), False)

    def test_filter_by_level(self):
        self.assertFalse(events.on_exception_info((None, 123, None), level='ignored'))
        self.assertIsNot(events.on_exception_info((None, 123, None), level='error'), False)

        self.assertFalse(events.on_message('hello world', level='ignored'))
        self.assertIsNot(events.on_message('hello world', level='error'), False)

        self.assertFalse(events.on_payload({}, level='ignored'))
        self.assertIsNot(events.on_message({}, level='error'), False)



import re

from rollbar.lib import events, filters

from rollbar.test import BaseTest


class CustomFiltersTest(BaseTest):
    def setUp(self):
        events.reset()
        filters.add_builtin_filters({})

    def test_ignore_by_setting_rollbar_ignore(self):
        class NotIgnoredByDefault(Exception):
            pass

        def _ignore_if_cruel_world_filter(exc_info, **kw):
            cls, exc, trace = exc_info
            if 'cruel world' in str(exc):
                exc._rollbar_ignore = True

            return exc_info

        events.add_exception_info_handler(_ignore_if_cruel_world_filter, pos=0)

        self.assertIsNot(events.on_exception_info((None, NotIgnoredByDefault('hello world'), None)), False)
        self.assertFalse(events.on_exception_info((None, NotIgnoredByDefault('hello cruel world'), None)))

    def test_ignore_messages_by_regex(self):
        regex = re.compile(r'cruel')

        def _ignore_cruel_world_substring(message, **kw):
            if regex.search(message):
                return False

            return message

        events.add_message_handler(_ignore_cruel_world_substring)

        self.assertFalse(events.on_message('hello cruel world'))
        self.assertIsNot(events.on_message('hello world'), False)

    def test_modify_payload(self):
        def _add_test_key(payload, **kw):
            payload['test'] = 333
            return payload

        events.add_payload_handler(_add_test_key)

        self.assertEqual(events.on_payload({'hello': 'world'}), {'hello': 'world', 'test': 333})



__all__ = ['LoggerMiddleware']

from fastapi import __version__

from rollbar.contrib.asgi.integration import integrate
from rollbar.contrib.starlette import LoggerMiddleware as StarletteLoggerMiddleware


@integrate(framework_name=f'fastapi {__version__}')
class LoggerMiddleware(StarletteLoggerMiddleware):
    ...



import sys
from array import array
from collections import deque

from rollbar import DEFAULT_LOCALS_SIZES, SETTINGS
from rollbar.lib import transforms
from rollbar.lib.transforms.shortener import ShortenerTransform
from rollbar.test import BaseTest


class TestClassWithAVeryVeryVeryVeryVeryVeryVeryLongName:
    pass


class KeyMemShortenerTransform(ShortenerTransform):
    """
    A shortener that just stores the keys.
    """
    keysUsed = []

    def default(self, o, key=None):
        self.keysUsed.append((key, o))
        return super(KeyMemShortenerTransform, self).default(o, key=key)


class ShortenerTransformTest(BaseTest):
    def setUp(self):
        self.shortener = ShortenerTransform(keys=[('shorten',)], **DEFAULT_LOCALS_SIZES)

    def test_shorten_string(self):
        original = 'x' * 120
        shortened = '{}...{}'.format('x'*48, 'x'*49)
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_long(self):
        original = 17955682733916468498414734863645002504519623752387
        shortened = '179556827339164684...5002504519623752387'
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_mapping(self):
        original = {
                'one': 'one',
                'two': 'two',
                'three': 'three',
                'four': 'four',
                'five': 'five',
                'six': 'six',
                'seven': 'seven',
                'eight': 'eight',
                'nine': 'nine',
                'ten': 'ten',
                'eleven': 'eleven'
            }
        shortened = {
                'one': 'one',
                'two': 'two',
                'three': 'three',
                'four': 'four',
                'five': 'five',
                'six': 'six',
                'seven': 'seven',
                'eight': 'eight',
                'nine': 'nine',
                'ten': 'ten',
            }
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_bytes(self):
        original = b'\x78' * 120
        shortened = b'\x78' * 100
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_list(self):
        original = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        shortened = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, '...']
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_tuple(self):
        original = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
        shortened = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, '...')
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_set(self):
        original = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11}
        shortened = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, '...'}
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_frozenset(self):
        original = frozenset([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        shortened = frozenset([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, '...'])
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_array(self):
        original = array('l', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        shortened = array('l', [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_deque(self):
        original = deque([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 15)
        shortened = deque([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 15)
        self.assertEqual(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_other(self):
        original = TestClassWithAVeryVeryVeryVeryVeryVeryVeryLongName()
        shortened = '<rollbar.test.test_shortener_transform.TestClas...'
        self.assertIn(shortened, self.shortener.default(original, ('shorten',)))
        self.assertEqual(original, self.shortener.default(original, ('nope',)))

    def test_shorten_object(self):
        data = {'request': {'POST': {i: i for i in range(12)}}}
        keys = [
                ('request', 'POST'),
                ('request', 'json'),
                ('body', 'request', 'POST'),
                ('body', 'request', 'json'),
                ]
        self.assertEqual(len(data['request']['POST']), 12)
        shortener = ShortenerTransform(keys=keys, **DEFAULT_LOCALS_SIZES)
        result = transforms.transform(data, shortener)
        self.assertEqual(type(result), dict)
        self.assertEqual(len(result['request']['POST']), 10)

    def test_shorten_frame(self):
        data = {
            'body': {
                'trace': {
                    'frames': [
                        {
                            "filename": "/path/to/app.py",
                            "lineno": 82,
                            "method": "sub_func",
                            "code": "extra(**kwargs)",
                            "keywordspec": "kwargs",
                            "locals": {
                                "kwargs": {
                                    "app": ["foo", "bar", "baz", "qux", "quux", "corge", "grault", "garply", "waldo",
                                            "fred", "plugh", "xyzzy", "thud"],
                                    "extra": {
                                        "request": "<class 'some.package.MyClass'>"
                                    }
                                },
                                "one": {
                                    "two": {
                                        "three": {
                                            "four": {
                                                "five": {
                                                    "six": {
                                                        "seven": 8,
                                                        "eight": "nine"
                                                    },
                                                    "ten": "Yep! this should still be here, but it is a little on the "
                                                           "long side, so we might want to cut it down a bit."
                                                }
                                            }
                                        }
                                    },
                                    "a": ["foo", "bar", "baz", "qux", 5, 6, 7, 8, 9, 10, 11, 12],
                                    "b": 14071504106566481658450568387453168916351054663,
                                    "app_id": 140715046161904,
                                    "bar": "im a bar",
                                }
                            }
                        }
                    ]
                }
            }
        }
        keys = [('body', 'trace', 'frames', '*', 'locals', '*')]
        shortener = ShortenerTransform(keys=keys, **DEFAULT_LOCALS_SIZES)
        result = transforms.transform(data, shortener)
        expected = {
            'body': {
                'trace': {
                    'frames': [
                        {
                            "filename": "/path/to/app.py",
                            "lineno": 82,
                            "method": "sub_func",
                            "code": "extra(**kwargs)",
                            "keywordspec": "kwargs",
                            "locals": {
                                "kwargs": {
                                    # Shortened
                                    "app": ['foo', 'bar', 'baz', 'qux', 'quux', 'corge', 'grault', 'garply', 'waldo',
                                           'fred', '...'],
                                    "extra": {
                                        "request": "<class 'some.package.MyClass'>"
                                    }
                                },
                                "one": {
                                    "two": {
                                        "three": {
                                            "four": {
                                                "five": {
                                                    "six": {'...': '...'},  # Dropped because it is past the maxlevel.
                                                    # Shortened
                                                    "ten": "Yep! this should still be here, but it is a litt...long "
                                                           "side, so we might want to cut it down a bit."
                                                }
                                            }
                                        }
                                    },
                                    "a": ['foo', 'bar', 'baz', 'qux', 5, 6, 7, 8, 9, 10, '...'],   # Shortened
                                    "b": '140715041065664816...7453168916351054663',  # Shortened
                                    "app_id": 140715046161904,
                                    "bar": "im a bar",
                                }
                            }
                        }
                    ]
                }
            }
        }

        self.assertEqual(result, expected)

    def test_breadth_first(self):
        obj = {
            "one": ["four", "five", 6, 7],
            "two": ("eight", "nine", "ten"),
            "three": {
                "eleven": 12,
                "thirteen": 14
            }
        }

        shortener_instance = KeyMemShortenerTransform(
            safe_repr=True,
            keys=[
                ('request', 'POST'),
                ('request', 'json'),
                ('body', 'request', 'POST'),
                ('body', 'request', 'json'),
            ],
            **SETTINGS['locals']['sizes']
        )

        transforms.transform(obj, [shortener_instance], key=())

        self.assertEqual(
            shortener_instance.keysUsed,
            [
                ((), {
                    "one": ["four", "five", 6, 7],
                    "two": ("eight", "nine", "ten"),
                    "three": {
                        "eleven": 12,
                        "thirteen": 14
                    }
                }),
                (("one",), ["four", "five", 6, 7]),
                (("one", 0), "four"),
                (("one", 1), "five"),
                (("one", 2), 6),
                (("one", 3), 7),
                (("two",), ("eight", "nine", "ten")),
                (("two", 0), "eight"),
                (("two", 1), "nine"),
                (("two", 2), "ten"),
                (("three",), {
                    "eleven": 12,
                    "thirteen": 14
                }),
                (("three", "eleven"), 12),
                (("three", "thirteen"), 14),
            ],
        )



from rollbar.lib import binary_type, string_types


from collections.abc import Mapping, Sequence, Set
from pathlib import Path


CIRCULAR = -1
DEFAULT = 0
MAPPING = 1
TUPLE = 2
NAMEDTUPLE = 3
LIST = 4
SET = 5
STRING = 6
PATH = 7


def get_type(obj):
    if isinstance(obj, (string_types, binary_type)):
        return STRING

    if isinstance(obj, Mapping):
        return MAPPING

    if isinstance(obj, tuple):
        if hasattr(obj, "_fields"):
            return NAMEDTUPLE

        return TUPLE

    if isinstance(obj, set):
        return SET

    if isinstance(obj, Sequence):
        return LIST

    if isinstance(obj, Path):
        return PATH

    return DEFAULT


__all__ = [
    "CIRCULAR",
    "DEFAULT",
    "MAPPING",
    "TUPLE",
    "NAMEDTUPLE",
    "LIST",
    "SET",
    "STRING",
    "PATH",
    "get_type",
]



import logging
import sys

from starlette import __version__
from starlette.requests import Request
from starlette.types import Receive, Scope, Send

import rollbar
from .requests import store_current_request
from rollbar.contrib.asgi import ReporterMiddleware as ASGIReporterMiddleware
from rollbar.contrib.asgi.integration import integrate
from rollbar.lib._async import RollbarAsyncError, try_report

log = logging.getLogger(__name__)


@integrate(framework_name=f'starlette {__version__}')
class ReporterMiddleware(ASGIReporterMiddleware):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            store_current_request(scope, receive)
            await self.app(scope, receive, send)
        except Exception:
            if scope['type'] == 'http':
                request = Request(scope, receive)

                # Consuming the request body in Starlette middleware is problematic.
                # See: https://github.com/encode/starlette/issues/495#issuecomment-494008175
                #
                # Uncomment lines below if you know the risks.
                #
                # Starlette requires the `python-multipart` package to parse the content
                # await request.body()
                # await request.form()

                exc_info = sys.exc_info()

                try:
                    await try_report(exc_info, request)
                except RollbarAsyncError:
                    log.warning(
                        'Failed to report asynchronously. Trying to report synchronously.'
                    )
                    rollbar.report_exc_info(exc_info, request)
            raise



from unittest import mock

from rollbar.test import BaseTest

try:
    from pyramid.request import Request

    PYRAMID_INSTALLED = True
except ImportError:
    PYRAMID_INSTALLED = False


if PYRAMID_INSTALLED:

    class PyramidMiddlewareTest(BaseTest):
        def test_catch_exception_in_the_wsgi_app(self):
            from rollbar.contrib.pyramid import RollbarMiddleware

            def wsgi_app(environ, start_resp):
                raise RuntimeError("oops")

            middleware = RollbarMiddleware({}, wsgi_app)

            with mock.patch("rollbar.report_exc_info") as mock_report:
                with self.assertRaises(RuntimeError):
                    middleware(environ={}, start_resp=lambda: None)

            self.assertEqual(mock_report.call_count, 1)

            args, kwargs = mock_report.call_args
            self.assertEqual(kwargs, {})

            exc_info, request = args

            exc_type, exc_value, exc_tb = exc_info
            self.assertEqual(exc_type, RuntimeError)
            self.assertIsInstance(exc_value, RuntimeError)

            self.assertIsInstance(request, Request)



#!/usr/bin/env python

# This example uses Uvicorn package that must be installed. However, it can be
# replaced with any other ASGI-compliant server.
#
# NOTE: Python 3.6 requires aiocontextvars package to be installed.
#
# Run: python app_global_request.py

import rollbar
import uvicorn

from rollbar.contrib.starlette import LoggerMiddleware
from starlette.applications import Starlette
from starlette.responses import JSONResponse

# Integrate Rollbar with Starlette application
app = Starlette()
app.add_middleware(LoggerMiddleware)  # should be added as the last middleware


async def get_user_agent():
    # Global access to the current request object
    request = rollbar.get_request()

    user_agent = request.headers['User-Agent']
    return user_agent


# $ curl -i http://localhost:8888
@app.route('/')
async def root(request):
    user_agent = await get_user_agent()
    return JSONResponse({'user-agent': user_agent})


if __name__ == '__main__':
    uvicorn.run(app, host='localhost', port=8888)



import logging
import sys

import rollbar
from .integration import IntegrationBase, integrate
from .types import ASGIApp, Receive, Scope, Send
from rollbar.lib._async import RollbarAsyncError, try_report

log = logging.getLogger(__name__)


@integrate(framework_name='asgi')
class ReporterMiddleware(IntegrationBase):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__()

        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            await self.app(scope, receive, send)
        except Exception:
            if scope['type'] == 'http':
                exc_info = sys.exc_info()

                try:
                    await try_report(exc_info)
                except RollbarAsyncError:
                    log.warning(
                        'Failed to report asynchronously. Trying to report synchronously.'
                    )
                    rollbar.report_exc_info(exc_info)
            raise



"""
Integration with Flask
"""

from flask import request
import rollbar


def report_exception(app, exception):
    rollbar.report_exc_info(request=request)


def _hook(request, data):
    data['framework'] = 'flask'

    if request:
        data['context'] = str(request.url_rule)

rollbar.BASE_DATA_HOOK = _hook



import logging
from pathlib import Path

from rollbar.lib import binary_type, string_types, circular_reference_label

# NOTE: Don't remove this line of code as it would cause a breaking change
# to the library's API. The items imported here were originally in this file
# but were moved to a new file for easier use elsewhere.
from rollbar.lib.type_info import (
    get_type,
    CIRCULAR,
    DEFAULT,
    MAPPING,
    TUPLE,
    NAMEDTUPLE,
    LIST,
    SET,
    STRING,
    PATH,
)


log = logging.getLogger(__name__)


def _noop_circular(a, **kw):
    return circular_reference_label(a, ref_key=kw.get("ref_key"))


def _noop(a, **_):
    return a


def _noop_tuple(a, **_):
    return tuple(a)


def _noop_namedtuple(a, **_):
    return a._make(a)


def _noop_list(a, **_):
    return list(a)


def _noop_set(a, **_):
    return set(a)


def _noop_mapping(a, **_):
    return dict(a)

def _noop_path(a, **_):
    return Path(a)


_default_handlers = {
    CIRCULAR: _noop_circular,
    DEFAULT: _noop,
    STRING: _noop,
    TUPLE: _noop_tuple,
    NAMEDTUPLE: _noop_namedtuple,
    LIST: _noop_list,
    SET: _noop_set,
    PATH: _noop_path,
    MAPPING: _noop_mapping,
}


def traverse(
    obj,
    key=(),
    string_handler=_default_handlers[STRING],
    tuple_handler=_default_handlers[TUPLE],
    namedtuple_handler=_default_handlers[NAMEDTUPLE],
    list_handler=_default_handlers[LIST],
    set_handler=_default_handlers[SET],
    mapping_handler=_default_handlers[MAPPING],
    path_handler=_default_handlers[PATH],
    default_handler=_default_handlers[DEFAULT],
    circular_reference_handler=_default_handlers[CIRCULAR],
    allowed_circular_reference_types=None,
    memo=None,
    depth_first=True,
    **custom_handlers
):
    memo = memo or {}
    obj_id = id(obj)
    obj_type = get_type(obj)

    ref_key = memo.get(obj_id)
    if ref_key:
        if not allowed_circular_reference_types or not isinstance(
            obj, allowed_circular_reference_types
        ):
            return circular_reference_handler(obj, key=key, ref_key=ref_key)

    memo[obj_id] = key

    kw = {
        "string_handler": string_handler,
        "tuple_handler": tuple_handler,
        "namedtuple_handler": namedtuple_handler,
        "list_handler": list_handler,
        "set_handler": set_handler,
        "mapping_handler": mapping_handler,
        "path_handler": path_handler,
        "default_handler": default_handler,
        "circular_reference_handler": circular_reference_handler,
        "allowed_circular_reference_types": allowed_circular_reference_types,
        "memo": memo,
        "depth_first": depth_first,
    }
    kw.update(custom_handlers)

    try:
        if obj_type is STRING:
            return string_handler(obj, key=key)
        elif obj_type is TUPLE:
            if depth_first:
                return tuple_handler(
                    tuple(
                        traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(obj)
                    ),
                    key=key,
                )
            # Breadth first
            return tuple(traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(tuple_handler(obj, key=key)))
        elif obj_type is NAMEDTUPLE:
            if depth_first:
                return namedtuple_handler(
                    obj._make(
                        traverse(v, key=key + (k,), **kw)
                        for k, v in obj._asdict().items()
                    ),
                    key=key,
                )
            # Breadth first
            return obj._make(traverse(v, key=key + (k,), **kw) for k, v in namedtuple_handler(obj, key=key)._asdict().items())
        elif obj_type is LIST:
            if depth_first:
                return list_handler(
                    [traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(obj)],
                    key=key,
                )
            # Breadth first
            return [traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(list_handler(obj, key=key))]
        elif obj_type is SET:
            if depth_first:
                return set_handler(
                    {traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(obj)},
                    key=key,
                )
            # Breadth first
            return {traverse(elem, key=key + (i,), **kw) for i, elem in enumerate(set_handler(obj, key=key))}
        elif obj_type is MAPPING:
            if depth_first:
                return mapping_handler(
                    {k: traverse(v, key=key + (k,), **kw) for k, v in obj.items()},
                    key=key,
                )
            # Breadth first
            return {k: traverse(v, key=key + (k,), **kw) for k, v in mapping_handler(obj, key=key).items()}
        elif obj_type is PATH:
            return path_handler(obj, key=key)
        elif obj_type is DEFAULT:
            for handler_type, handler in custom_handlers.items():
                if isinstance(obj, handler_type):
                    return handler(obj, key=key)
    except:
        # use the default handler for unknown object types
        log.debug(
            "Exception while traversing object using type-specific "
            "handler. Switching to default handler.",
            exc_info=True,
        )

    return default_handler(obj, key=key)


__all__ = ["traverse"]



from rollbar.lib.transform import Transform
from rollbar.lib.traverse import traverse

from rollbar.test import BaseTest


class NamedTuple(tuple):
    """
    Modeled after NamedTuple and KeyedTuple from SQLAlchemy 0.7 and 0.8.
    """

    def __new__(cls, vals, labels=None):
        t = tuple.__new__(cls, vals)
        if labels:
            t.__dict__.update(zip(labels, vals))
            t._labels = labels
        return t

    def keys(self):
        return [l for l in self._labels if l is not None]


class KeyMemTransform(Transform):
    """
    A transform that just stores the keys.
    """
    keys = []

    def default(self, o, key=None):
        self.keys.append((key, o))
        return o


class RollbarTraverseTest(BaseTest):
    """
    Objects that appear to be a namedtuple, like SQLAlchemy's KeyedTuple,
    will cause an Exception while identifying them if they don't implement
    the _make method.
    """

    def setUp(self):
        self.tuple = NamedTuple((1, 2, 3), labels=["one", "two", "three"])

    def test_base_case(self):
        self.assertEqual(traverse(self.tuple), (1, 2, 3))

    def test_bad_object(self):
        setattr(self.tuple, "_fields", "not quite a named tuple")
        self.assertEqual(traverse(self.tuple), (1, 2, 3))

    def test_depth_first(self):
        obj = {
            "one": ["four", "five", 6, 7],
            "two": ("eight", "nine", "ten"),
            "three": {
                "eleven": 12,
                "thirteen": 14
            }
        }
        transform = KeyMemTransform()
        transform.keys = []

        traverse(
            obj,
            key=(),
            string_handler=transform.default,
            tuple_handler=transform.default,
            namedtuple_handler=transform.default,
            list_handler=transform.default,
            set_handler=transform.default,
            mapping_handler=transform.default,
            path_handler=transform.default,
            default_handler=transform.default,
            circular_reference_handler=transform.default,
            allowed_circular_reference_types=transform.default,
        )

        self.assertEqual(
            transform.keys,
            [
                (("one", 0), "four"),
                (("one", 1), "five"),
                (("one", 2), 6),
                (("one", 3), 7),
                (("one",), ["four", "five", 6, 7]),
                (("two", 0), "eight"),
                (("two", 1), "nine"),
                (("two", 2), "ten"),
                (("two",), ("eight", "nine", "ten")),
                (("three", "eleven"), 12),
                (("three", "thirteen"), 14),
                (("three",), {
                    "eleven": 12,
                    "thirteen": 14
                }),
                ((), {
                    "one": ["four", "five", 6, 7],
                    "two": ("eight", "nine", "ten"),
                    "three": {
                        "eleven": 12,
                        "thirteen": 14
                    }
                }),
            ],
        )

    def test_breadth_first(self):
        obj = {
            "one": ["four", "five", 6, 7],
            "two": ("eight", "nine", "ten"),
            "three": {
                "eleven": 12,
                "thirteen": 14
            }
        }
        transform = KeyMemTransform()
        transform.keys = []

        traverse(
            obj,
            key=(),
            string_handler=transform.default,
            tuple_handler=transform.default,
            namedtuple_handler=transform.default,
            list_handler=transform.default,
            set_handler=transform.default,
            mapping_handler=transform.default,
            path_handler=transform.default,
            default_handler=transform.default,
            circular_reference_handler=transform.default,
            allowed_circular_reference_types=transform.default,
            depth_first=False,
        )

        self.assertEqual(
            transform.keys,
            [
                ((), {
                    "one": ["four", "five", 6, 7],
                    "two": ("eight", "nine", "ten"),
                    "three": {
                        "eleven": 12,
                        "thirteen": 14
                    }
                }),
                (("one",), ["four", "five", 6, 7]),
                (("one", 0), "four"),
                (("one", 1), "five"),
                (("one", 2), 6),
                (("one", 3), 7),
                (("two",), ("eight", "nine", "ten")),
                (("two", 0), "eight"),
                (("two", 1), "nine"),
                (("two", 2), "ten"),
                (("three",), {
                    "eleven": 12,
                    "thirteen": 14
                }),
                (("three", "eleven"), 12),
                (("three", "thirteen"), 14),
            ],
        )



import functools
import logging
from typing import Union

import fastapi
from fastapi import APIRouter, FastAPI

from . import ReporterMiddleware as FastAPIReporterMiddleware
from rollbar.contrib.starlette import ReporterMiddleware as StarletteReporterMiddleware
from rollbar.contrib.asgi import ReporterMiddleware as ASGIReporterMiddleware

log = logging.getLogger(__name__)


class FastAPIVersionError(Exception):
    def __init__(self, version, reason=''):
        err_msg = f'FastAPI {version}+ is required'
        if reason:
            err_msg += f' {reason}'

        log.error(err_msg)
        return super().__init__(err_msg)


def is_current_version_higher_or_equal(current_version, min_version):
    """
    Compare two version strings and return True if the current version is higher or equal to the minimum version.

    Note: This function only compares the release segment of the version string.
    """
    def parse_version(version):
        """Parse the release segment of a version string into a list of strings."""
        parsed = ['']
        current_segment = 0
        for c in version:
            if c.isdigit():
                parsed[current_segment] += c
            elif c == '.':
                current_segment += 1
                parsed.append('')
            else:
                break
        if parsed[-1] == '':
            parsed.pop()
        return parsed

    current = tuple(map(int, parse_version(current_version)))
    minimum = tuple(map(int, parse_version(min_version)))
    return current >= minimum


class fastapi_min_version:
    def __init__(self, min_version):
        self.min_version = min_version

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not is_current_version_higher_or_equal(
                fastapi.__version__,
                self.min_version,
            ):
                raise FastAPIVersionError(
                    self.min_version, reason=f'to use {func.__name__}() function'
                )

            return func(*args, **kwargs)

        return wrapper


def get_installed_middlewares(app):
    candidates = (
        FastAPIReporterMiddleware,
        StarletteReporterMiddleware,
        ASGIReporterMiddleware,
    )

    middlewares = []

    if hasattr(app, 'user_middleware'):  # FastAPI v0.51.0+
        middlewares = [
            middleware.cls
            for middleware in app.user_middleware
            if middleware.cls in candidates
        ]
    elif hasattr(app, 'error_middleware'):
        middleware = app.error_middleware

        while hasattr(middleware, 'app'):
            if isinstance(middleware, candidates):
                middlewares.append(middleware)
            middleware = middleware.app

        middlewares = [middleware.__class__ for middleware in middlewares]

    return middlewares


def has_bare_routing(app_or_router: Union[FastAPI, APIRouter]):
    if not isinstance(app_or_router, (FastAPI, APIRouter)):
        return False

    urls = [
        getattr(app_or_router, 'openapi_url', None),
        getattr(app_or_router, 'docs_url', None),
        getattr(app_or_router, 'redoc_url', None),
        getattr(app_or_router, 'swagger_ui_oauth2_redirect_url', None),
    ]

    for route in app_or_router.routes:
        if route is None or route.path in urls:
            continue
        return False

    return True



import os
import sys

import django
from django.conf import settings


# use settings compatible with the installed Django version
# v1.10+ requires MIDDLEWARE keyname
# older versions require MIDDLEWARE_CLASSES keyname

ROLLBAR_CONFIG = {
    'access_token': 'POST_SERVER_ITEM_ACCESS_TOKEN',
    'environment': 'development',
    'branch': 'master',
    'root': os.getcwd()
}

MIDDLEWARE_CONFIG = (
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',

    # Rollbar middleware
    'rollbar.contrib.django.middleware.RollbarNotifierMiddleware',
)

if django.VERSION >= (1, 10):
    settings.configure(
        DEBUG=True,
        SECRET_KEY='thisisthesecretkey',
        ROOT_URLCONF=__name__,
        ROLLBAR = ROLLBAR_CONFIG,
        MIDDLEWARE = MIDDLEWARE_CONFIG,
    )
else:
    settings.configure(
        DEBUG=True,
        SECRET_KEY='thisisthesecretkey',
        ROOT_URLCONF=__name__,
        ROLLBAR = ROLLBAR_CONFIG,
        MIDDLEWARE_CLASSES = MIDDLEWARE_CONFIG,
    )

from django.conf.urls import url
from django.http import HttpResponse


def index(request):
    return HttpResponse('Hello World')

def error(request):
    foo()
    return HttpResponse('You shouldn\'t be seeing this')


urlpatterns = (
    url(r'^$', index),
    url(r'^error$', error),
)

if __name__ == "__main__":
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)




# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.
#
# From https://twistedmatrix.com/documents/current/_downloads/simpleserv.py

# NOTE: pyrollbar requires both `Twisted` and `treq` packages to be installed

from twisted.internet import reactor, protocol

import rollbar


def bar(p):
    # These local variables will be sent to Rollbar and available in the UI
    a = 33
    b = a * 5
    baz()


def foo():
    hello = 'world'
    bar(hello)


class Echo(protocol.Protocol):
    """This is just about the simplest possible protocol"""

    def dataReceived(self, data):
        "As soon as any data is received, write it back."

        # Cause an uncaught exception to be sent to Rollbar
        foo()
        self.transport.write(data)


def main():
    rollbar.init('ACCESS_TOKEN', environment='test', handler='twisted')

    """This runs the protocol on port 8000"""
    factory = protocol.ServerFactory()
    factory.protocol = Echo
    reactor.listenTCP(8000, factory)
    reactor.run()


# this only runs if the module was *not* imported
if __name__ == '__main__':
    main()



from collections.abc import Iterable

from rollbar.lib import (
    binary_type,
    string_types,
    number_types,
    traverse,
)
# NOTE: Don't remove this import, it would cause a breaking change to the library's API.
# The `Transform` class was moved out of this file to prevent a cyclical dependency issue.
from rollbar.lib.transform import Transform
from rollbar.lib.transforms.batched import BatchedTransform

_ALLOWED_CIRCULAR_REFERENCE_TYPES = [binary_type, bool, type(None)]

if isinstance(string_types, tuple):
    _ALLOWED_CIRCULAR_REFERENCE_TYPES.extend(string_types)
else:
    _ALLOWED_CIRCULAR_REFERENCE_TYPES.append(string_types)

if isinstance(number_types, tuple):
    _ALLOWED_CIRCULAR_REFERENCE_TYPES.extend(number_types)
else:
    _ALLOWED_CIRCULAR_REFERENCE_TYPES.append(number_types)

_ALLOWED_CIRCULAR_REFERENCE_TYPES = tuple(_ALLOWED_CIRCULAR_REFERENCE_TYPES)


def transform(obj, transforms, key=None, batch_transforms=False):
    if isinstance(transforms, Transform):
        transforms = [transforms]

    if batch_transforms:
        transforms = [BatchedTransform(transforms)]

    for transform in transforms:
        if not isinstance(transform, Transform):
            continue
        obj = _transform(obj, transform, key=key)

    return obj


def _transform(obj, transform, key=None):
    key = key or ()

    def do_transform(type_name, val, key=None, **kw):
        fn = getattr(transform, "transform_%s" % type_name, transform.transform_custom)
        val = fn(val, key=key, **kw)

        return val

    def string_handler(s, key=None):
        if isinstance(s, bytes):
            return do_transform("bytes", s, key=key)
        elif isinstance(s, str):
            return do_transform("unicode", s, key=key)

    def default_handler(o, key=None):
        if isinstance(o, bool):
            return do_transform("boolean", o, key=key)

        # There is a quirk in the current version (1.1.6) of the enum
        # backport enum34 which causes it to not have the same
        # behavior as Python 3.4+. One way to identify IntEnums is that
        # they are instances of numbers but not number types.
        if isinstance(o, number_types):
            if type(o) not in number_types:
                return do_transform("custom", o, key=key)
            else:
                return do_transform("number", o, key=key)

        return do_transform("custom", o, key=key)

    handlers = {
        "string_handler": string_handler,
        "tuple_handler": lambda o, key=None: do_transform("tuple", o, key=key),
        "namedtuple_handler": lambda o, key=None: do_transform(
            "namedtuple", o, key=key
        ),
        "list_handler": lambda o, key=None: do_transform("list", o, key=key),
        "set_handler": lambda o, key=None: do_transform("set", o, key=key),
        "mapping_handler": lambda o, key=None: do_transform("dict", o, key=key),
        "path_handler": lambda o, key=None: do_transform("path", o, key=key),
        "circular_reference_handler": lambda o, key=None, ref_key=None: do_transform(
            "circular_reference", o, key=key, ref_key=ref_key
        ),
        "default_handler": default_handler,
        "allowed_circular_reference_types": _ALLOWED_CIRCULAR_REFERENCE_TYPES,
    }

    return traverse.traverse(obj, key=key, depth_first=transform.depth_first, **handlers)


__all__ = ["transform", "Transform"]



from array import array
import collections
import itertools
import reprlib

from collections.abc import Mapping
from typing import Union, Tuple

from rollbar.lib import (
    integer_types, key_in, key_depth, sequence_types,
    string_types)
from rollbar.lib.transform import Transform


_type_name_mapping = {
    'string': string_types,
    'long': integer_types,
    'mapping': Mapping,
    'list': list,
    'tuple': tuple,
    'set': set,
    'frozenset': frozenset,
    'array': array,
    'deque': collections.deque,
    'other': None
}


def _max_left_right(max_len: int, seperator_len: int) -> Tuple[int, int]:
    left = max(0, (max_len-seperator_len)//2)
    right = max(0, max_len-seperator_len-left)
    return left, right


def shorten_array(obj: array, max_len: int) -> array:
    if len(obj) <= max_len:
        return obj

    return obj[:max_len]


def shorten_bytes(obj: bytes, max_len: int) -> bytes:
    if len(obj) <= max_len:
        return obj

    return obj[:max_len]


def shorten_deque(obj: collections.deque, max_len: int) -> collections.deque:
    if len(obj) <= max_len:
        return obj

    return collections.deque(itertools.islice(obj, max_len))


def shorten_frozenset(obj: frozenset, max_len: int) -> frozenset:
    if len(obj) <= max_len:
        return obj

    return frozenset([elem for i, elem in enumerate(obj) if i < max_len] + ['...'])


def shorten_int(obj: int, max_len: int) -> Union[int, str]:
    s = repr(obj)
    if len(s) <= max_len:
        return obj

    left, right = _max_left_right(max_len, 3)
    return s[:left] + '...' + s[len(s)-right:]


def shorten_list(obj: list, max_len: int) -> list:
    if len(obj) <= max_len:
        return obj

    return obj[:max_len] + ['...']


def shorten_mapping(obj: Union[dict, Mapping], max_keys: int) -> dict:
    if len(obj) <= max_keys:
        return obj

    return {k: obj[k] for k in itertools.islice(obj.keys(), max_keys)}


def shorten_set(obj: set, max_len: int) -> set:
    if len(obj) <= max_len:
        return obj

    return set([elem for i, elem in enumerate(obj) if i < max_len] + ['...'])


def shorten_string(obj: str, max_len: int) -> str:
    if len(obj) <= max_len:
        return obj

    left, right = _max_left_right(max_len, 3)
    return obj[:left] + '...' + obj[len(obj)-right:]


def shorten_tuple(obj: tuple, max_len: int) -> tuple:
    if len(obj) <= max_len:
        return obj

    return obj[:max_len] + ('...',)



class ShortenerTransform(Transform):
    depth_first = False

    def __init__(self, safe_repr=True, keys=None, **sizes):
        super(ShortenerTransform, self).__init__()
        self.safe_repr = safe_repr
        self.keys = keys
        self._repr = reprlib.Repr()

        for name, size in sizes.items():
            setattr(self._repr, name, size)

    def _get_max_size(self, obj):
        for name, _type in _type_name_mapping.items():
            # Special case for dicts since we are using collections.abc.Mapping
            # to provide better type checking for dict-like objects
            if name == 'mapping':
                name = 'dict'

            if _type and isinstance(obj, _type):
                return getattr(self._repr, 'max%s' % name)

        return self._repr.maxother

    def _shorten(self, val):
        max_size = self._get_max_size(val)

        if isinstance(val, array):
            return shorten_array(val, max_size)
        if isinstance(val, bytes):
            return shorten_bytes(val, max_size)
        if isinstance(val, collections.deque):
            return shorten_deque(val, max_size)
        if isinstance(val, (dict, Mapping)):
            return shorten_mapping(val, max_size)
        if isinstance(val, float):
            return val
        if isinstance(val, frozenset):
            return shorten_frozenset(val, max_size)
        if isinstance(val, int):
            return shorten_int(val, max_size)
        if isinstance(val, list):
            return shorten_list(val, max_size)
        if isinstance(val, set):
            return shorten_set(val, max_size)
        if isinstance(val, str):
            return shorten_string(val, max_size)
        if isinstance(val, tuple):
            return shorten_tuple(val, max_size)

        return self._shorten_other(val)

    def _shorten_other(self, obj):
        if obj is None:
            return None

        if self.safe_repr:
            obj = str(obj)

        return self._repr.repr(obj)

    def _should_shorten(self, val, key):
        if not key:
            return False

        return key_in(key, self.keys)

    def _should_drop(self, val, key) -> bool:
        if not key:
            return False

        max_depth = key_depth(key, self.keys)
        if max_depth == 0:
            return False

        return (max_depth + self._repr.maxlevel) <= len(key)

    def default(self, o, key=None):
        if self._should_drop(o, key):
            if isinstance(o, (dict, Mapping)):
                return {'...': '...'}
            if isinstance(o, sequence_types):
                return ['...']

        if self._should_shorten(o, key):
            return self._shorten(o)

        return super(ShortenerTransform, self).default(o, key=key)


__all__ = ['ShortenerTransform']



import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor

_pool = None  # type: ThreadPoolExecutor|None

log = logging.getLogger(__name__)


def init_pool(max_workers):
    """
    Creates the thread pool with the max workers.

    :type max_workers: int|None
    :param max_workers: If max_workers is None it will use the logic from the standard library to calculate the number
                        of threads. However, we ported the logic from Python 3.5 to earlier versions.
    """
    if max_workers is None and sys.version_info < (3, 5):
        max_workers = (os.cpu_count() or 1) * 5

    global _pool
    _pool = ThreadPoolExecutor(max_workers)


def submit(worker, payload_str, access_token):
    """
    Submit a new task to the thread pool.

    :type worker: function
    :type payload_str: str
    :type access_token: str
    """
    global _pool
    if _pool is None:
        log.warning('pyrollbar: Thead pool not initialized. Please ensure init_pool() is called prior to submit().')
        return
    _pool.submit(worker, payload_str, access_token)



r"""
django-rollbar middleware

There are two options for installing the Rollbar middleware. Both options
require modifying your settings.py file.

The first option is to use
'rollbar.contrib.django.middleware.RollbarNotifierMiddleware' which will
report all exceptions to Rollbar including 404s. This middlware should be
placed as the last item in your middleware list which is:
    * MIDDLEWARE_CLASSES in Django 1.9 and earlier
    * MIDDLEWARE in Django 1.10 and up

The other option is two use the two separate middlewares:
    * 'rollbar.contrib.django.middleware.RollbarNotifierMiddlewareExcluding404'
    * 'rollbar.contrib.django.middleware.RollbarNotifierMiddlewareOnly404'
The Excluding404 middleware should be placed as the last item in your middleware
list, and the Only404 middleware should be placed as the first item in your
middleware list. This allows 404s to be processed by your other middlewares
before sendind an item to Rollbar. Therefore if you handle the 404 differently
in a way that returns a response early you won't end up with a Rollbar item.

Regardless of which method you use, you also should add a section to settings.py
like this:

ROLLBAR = {
    'access_token': 'tokengoeshere',
}

This can be used for passing configuration options to Rollbar. Additionally,
you can use the key 'ignorable_404_urls' to set an iterable of regular expression
patterns to use to determine whether a 404 exception should be ignored based
on the full url path for the request. For example,

import re
ROLLBAR = {
    'access_token': 'YOUR_TOKEN',
    'ignorable_404_urls': (
        re.compile(r'/index\.php'),
        re.compile('/foobar'),
    ),
}

To get more control of middleware and enrich it with custom data
you can subclass any of the middleware classes described above
and optionally override the methods:
    def get_extra_data(self, request, exc):
        ''' May be defined.  Must return a dict or None. Use it to put some custom extra data on rollbar event. '''
        return

    def get_payload_data(self, request, exc):
        ''' May be defined.  Must return a dict or None. Use it to put some custom payload data on rollbar event. '''
        return
You would then insert your custom subclass into your middleware
configuration in the same place as the base class as described above.
For example:

1. create a 'middleware.py' file on your project (name is up to you)
2. import the rollbar default middleware: 'from rollbar.contrib.django.middleware import RollbarNotifierMiddleware'
3. create your own middleware like this:
class CustomRollbarNotifierMiddleware(RollbarNotifierMiddleware):
    def get_extra_data(self, request, exc):
        ''' May be defined.  Must return a dict or None. Use it to put some custom extra data on rollbar event. '''
        return

    def get_payload_data(self, request, exc):
        ''' May be defined.  Must return a dict or None. Use it to put some custom payload data on rollbar event. '''
        return

4. add 'path.to.your.CustomRollbarNotifierMiddleware' in your settings.py to
    a. MIDDLEWARE_CLASSES in Django 1.9 and earlier
    b. MIDDLEWARE in Django 1.10 and up
5. add a section like this in your settings.py:
ROLLBAR = {
    'access_token': 'tokengoeshere',
}

See README.rst for full installation and configuration instructions.
"""

import logging
import sys

import rollbar

from django.core.exceptions import MiddlewareNotUsed
from django.conf import settings
from django.http import Http404

try:
    from django.urls import resolve
except ImportError:
    from django.core.urlresolvers import resolve

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    from rollbar.contrib.django.utils import MiddlewareMixin

log = logging.getLogger(__name__)


DEFAULTS = {
    'web_base': 'https://rollbar.com',
    'enabled': True,
    'patch_debugview': True,
    'exception_level_filters': [
        (Http404, 'warning')
    ]
}


def _patch_debugview(rollbar_web_base):
    try:
        from django.views import debug
    except ImportError:
        return

    if rollbar_web_base.endswith('/'):
        rollbar_web_base = rollbar_web_base[:-1]

    # modify the TECHNICAL_500_TEMPLATE
    new_data = """
{% if view_in_rollbar_url %}
  <h3 style="margin-bottom:15px;"><a href="{{ view_in_rollbar_url }}" target="_blank">View in Rollbar</a></h3>
{% endif %}
    """

    insert_before = '<table class="meta">'
    replacement = new_data + insert_before

    if hasattr(debug, 'TECHNICAL_500_TEMPLATE'):
        if new_data in debug.TECHNICAL_500_TEMPLATE:
            return
        debug.TECHNICAL_500_TEMPLATE = debug.TECHNICAL_500_TEMPLATE.replace(insert_before, replacement, 1)
    elif hasattr(debug, 'CURRENT_DIR'):
        # patch ExceptionReporter.get_traceback_html if this version of Django is using
        # the file system templates rather than the ones in code
        # This code comes from:
        # https://github.com/django/django/blob/d79cf1e9e2887aa12567c8f27e384195253cb847/django/views/debug.py#L329,L334
        # There are theoretical issues with the code below, for example t.render could throw because
        # t might be None, but this is the code from Django
        from pathlib import Path
        from django.template import Context
        def new_get_traceback_html(exception_reporter):
            """Return HTML version of debug 500 HTTP error page."""
            with Path(debug.CURRENT_DIR, 'templates', 'technical_500.html').open() as fh:
                template_string = fh.read()
                template_string = template_string.replace(insert_before, replacement, 1)
                t = debug.DEBUG_ENGINE.from_string(template_string)
            c = Context(exception_reporter.get_traceback_data(), use_l10n=False)
            return t.render(c)
        debug.ExceptionReporter.get_traceback_html = new_get_traceback_html
    else:
        # patch ExceptionReporter.get_traceback_html for Django versions 4.0+
        def new_get_traceback_html(self):
            """Return HTML version of debug 500 HTTP error page."""
            with self.html_template_path.open(encoding='utf-8') as fh:
                t = debug.DEBUG_ENGINE.from_string(fh.read())
            c = Context(self.get_traceback_data(), use_l10n=False)
            return t.render(c)

    if hasattr(debug.ExceptionReporter, '__rollbar__patched'):
        return

    # patch ExceptionReporter.get_traceback_data
    old_get_traceback_data = debug.ExceptionReporter.get_traceback_data
    def new_get_traceback_data(exception_reporter):
        data = old_get_traceback_data(exception_reporter)
        try:
            item_uuid = exception_reporter.request.META.get('rollbar.uuid')
            if item_uuid:
                url = '%s/item/uuid/?uuid=%s' % (rollbar_web_base, item_uuid)
                data['view_in_rollbar_url'] = url
        except:
            log.exception("Exception while adding view-in-rollbar link to technical_500_template.")
        return data
    debug.ExceptionReporter.get_traceback_data = new_get_traceback_data
    debug.ExceptionReporter.__rollbar__patched = True



def _should_ignore_404(url):
    url_patterns = getattr(settings, 'ROLLBAR', {}).get('ignorable_404_urls', ())
    return any(p.search(url) for p in url_patterns)

def _apply_sensitive_post_params(request):
    sensitive_post_parameters = getattr(
        request, "sensitive_post_parameters", []
    )
    if not sensitive_post_parameters:
        return
    mutable = request.POST._mutable
    request.POST._mutable = True

    if sensitive_post_parameters == "__ALL__":
        for param in request.POST:
            request.POST[param] = "******"
        return

    for param in sensitive_post_parameters:
        if param in request.POST:
            request.POST[param] = "******"
    request.POST._mutable = mutable

class RollbarNotifierMiddleware(MiddlewareMixin):
    def __init__(self, get_response=None):
        super(RollbarNotifierMiddleware, self).__init__(get_response)

        self.settings = getattr(settings, 'ROLLBAR', {})
        if not self.settings.get('access_token'):
            raise MiddlewareNotUsed

        if not self._get_setting('enabled'):
            raise MiddlewareNotUsed

        self._ensure_log_handler()

        kw = self.settings.copy()
        access_token = kw.pop('access_token')
        environment = kw.pop('environment', 'development' if settings.DEBUG else 'production')
        kw.setdefault('exception_level_filters', DEFAULTS['exception_level_filters'])

        # ignorable_404_urls is only relevant for this middleware not as an argument to init
        kw.pop('ignorable_404_urls', None)

        rollbar.init(access_token, environment, **kw)

        def hook(request, data):
            try:
                # try django 1.5 method for getting url_name
                url_name = request.resolver_match.url_name
            except:
                # fallback to older method
                try:
                    url_name = resolve(request.path_info).url_name
                except:
                    url_name = None

            if url_name:
                data['context'] = url_name

            data['framework'] = 'django'

            if request and hasattr(request, 'META'):
                request.META['rollbar.uuid'] = data['uuid']

        rollbar.BASE_DATA_HOOK = hook

        # monkeypatch debug module
        if self._get_setting('patch_debugview'):
            try:
                _patch_debugview(self._get_setting('web_base'))
            except Exception as e:
                log.error(
                    "Rollbar - unable to monkeypatch debugview to add 'View in Rollbar' link."
                    " To disable, set `ROLLBAR['patch_debugview'] = False` in settings.py."
                    " Exception was: %r", e
                )

    def _ensure_log_handler(self):
        """
        If there's no log configuration, set up a default handler.
        """
        if log.handlers:
            return
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s')
        handler.setFormatter(formatter)
        log.addHandler(handler)

    def _get_setting(self, name, default=None):
        try:
            return self.settings[name]
        except KeyError:
            if name in DEFAULTS:
                default_val = DEFAULTS[name]
                if hasattr(default_val, '__call__'):
                    return default_val()
                return default_val
            return default

    def get_extra_data(self, request, exc):
        return

    def get_payload_data(self, request, exc):
        return

    def process_response(self, request, response):
        return response

    def process_exception(self, request, exc):
        if isinstance(exc, Http404) and _should_ignore_404(request.get_full_path()):
            return
        _apply_sensitive_post_params(request)

        rollbar.report_exc_info(
            sys.exc_info(),
            request,
            extra_data=self.get_extra_data(request, exc),
            payload_data=self.get_payload_data(request, exc),
        )


class RollbarNotifierMiddlewareOnly404(MiddlewareMixin):
    def get_extra_data(self, request, exc):
        return

    def get_payload_data(self, request, exc):
        return

    def process_response(self, request, response):
        if response.status_code != 404:
            return response

        if _should_ignore_404(request.get_full_path()):
            return response

        try:
            if hasattr(request, '_rollbar_notifier_original_http404_exc_info'):
                exc_type, exc_value, exc_traceback = request._rollbar_notifier_original_http404_exc_info
                if exc_value is None:
                    exc_value = Http404()
                raise exc_value.with_traceback(exc_traceback)
            else:
                raise Http404()
        except Exception as exc:
            _apply_sensitive_post_params(request)
            rollbar.report_exc_info(
                sys.exc_info(),
                request,
                extra_data=self.get_extra_data(request, exc),
                payload_data=self.get_payload_data(request, exc),
            )
        return response


class RollbarNotifierMiddlewareExcluding404(RollbarNotifierMiddleware):
    def process_exception(self, request, exc):
        if isinstance(exc, Http404):
            request._rollbar_notifier_original_http404_exc_info = sys.exc_info()
        else:
            super(RollbarNotifierMiddlewareExcluding404, self).process_exception(request, exc)



import collections
import base64
import copy
import enum
import sys

from collections.abc import Mapping

from rollbar.lib import transforms
from rollbar.lib.transforms.serializable import SerializableTransform

from rollbar.test import BaseTest, SNOWMAN_UNICODE

SNOWMAN_LEN = len(SNOWMAN_UNICODE)


# This base64 encoded string contains bytes that do not
# convert to utf-8 data
invalid_b64 = b'CuX2JKuXuLVtJ6l1s7DeeQ=='

invalid = base64.b64decode(invalid_b64)
undecodable_repr = f'<Undecodable type:({"bytes"}) base64:({invalid_b64.decode("ascii")})>'


class SerializableTransformTest(BaseTest):
    def _assertSerialized(self, start, expected, safe_repr=True, safelist=None, skip_id_check=False):
        serializable = SerializableTransform(safe_repr=safe_repr, safelist_types=safelist)
        result = transforms.transform(start, serializable)

        """
        #print start
        print result
        print expected
        """

        if not skip_id_check:
            self.assertNotEqual(id(result), id(expected))

        self.assertEqual(type(expected), type(result))

        if isinstance(result, Mapping):
            self.assertDictEqual(result, expected)
        elif isinstance(result, tuple):
            self.assertTupleEqual(result, expected)
        elif isinstance(result, (list, set)):
            self.assertListEqual(result, expected)
        else:
            self.assertEqual(result, expected)

    def test_simple_dict(self):
        start = {
            'hello': 'world',
            '1': 2,
        }
        expected = copy.deepcopy(start)
        self._assertSerialized(start, expected)

    def test_enum(self):
        class EnumTest(enum.Enum):
            one = 1
            two = 2

        start = EnumTest.one
        expected = "<enum 'EnumTest'>"
        self._assertSerialized(start, expected)

    def test_enum_no_safe_repr(self):
        class EnumTest(enum.Enum):
            one = 1
            two = 2

        start = EnumTest.one
        expected = '<EnumTest.one: 1>'
        self._assertSerialized(start, expected, safe_repr=False)

    def test_int_enum(self):
        class EnumTest(enum.IntEnum):
            one = 1
            two = 2

        start = EnumTest.one
        expected = "<enum 'EnumTest'>"
        self._assertSerialized(start, expected)

    def test_int_enum_no_safe_repr(self):
        class EnumTest(enum.IntEnum):
            one = 1
            two = 2

        start = EnumTest.one
        expected = '<EnumTest.one: 1>'
        self._assertSerialized(start, expected, safe_repr=False)

    def test_encode_dict_with_invalid_utf8(self):
        start = {
            'invalid': invalid
        }
        expected = copy.copy(start)
        expected['invalid'] = undecodable_repr
        self._assertSerialized(start, expected)

    def test_encode_utf8(self):
        start = invalid
        expected = undecodable_repr
        self._assertSerialized(start, expected)

    def test_encode_None(self):
        start = None
        expected = None
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_float(self):
        start = 3.14
        expected = 3.14
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_float_nan(self):
        start = float('nan')
        expected = '<NaN>'
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_float_infinity(self):
        start = float('inf')
        expected = '<Infinity>'
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_float_neg_infinity(self):
        start = float('-inf')
        expected = '<NegativeInfinity>'
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_int(self):
        start = 33
        expected = 33
        self._assertSerialized(start, expected, skip_id_check=True)

    def test_encode_empty_tuple(self):
        start = ()
        expected = ()
        
        skip_id_check = False
        # different behavior in 3.11
        if sys.version_info >= (3, 11):
            skip_id_check = True
        
        self._assertSerialized(start, expected, skip_id_check=skip_id_check)

    def test_encode_empty_list(self):
        start = []
        expected = []
        self._assertSerialized(start, expected)

    def test_encode_empty_dict(self):
        start = {}
        expected = {}
        self._assertSerialized(start, expected)

    def test_encode_namedtuple(self):
        MyType = collections.namedtuple('MyType', ('field_1', 'field_2'))
        nt = MyType(field_1='this is field 1', field_2=invalid)

        start = nt
        expected = "<MyType(field_1='this is field 1', field_2='%s')>" % undecodable_repr

        self._assertSerialized(start, expected)

    def test_encode_tuple_with_bytes(self):
        start = ('hello', 'world', invalid)
        expected = list(start)
        expected[2] = undecodable_repr
        self._assertSerialized(start, tuple(expected))

    def test_encode_list_with_bytes(self):
        start = ['hello', 'world', invalid]
        expected = list(start)
        expected[2] = undecodable_repr
        self._assertSerialized(start, expected)

    def test_encode_dict_with_bytes(self):
        start = {'hello': 'world', 'invalid': invalid}
        expected = copy.deepcopy(start)
        expected['invalid'] = undecodable_repr
        self._assertSerialized(start, expected)

    def test_encode_dict_with_bytes_key(self):
        start = {'hello': 'world', invalid: 'works?'}
        expected = copy.deepcopy(start)
        expected[undecodable_repr] = 'works?'
        del expected[invalid]
        self._assertSerialized(start, expected)

    def test_encode_with_custom_repr_no_safelist(self):
        class CustomRepr(object):
            def __repr__(self):
                return 'hello'

        start = {'hello': 'world', 'custom': CustomRepr()}
        expected = copy.deepcopy(start)
        expected['custom'] = str(CustomRepr)
        self._assertSerialized(start, expected)

    def test_encode_with_custom_repr_no_safelist_no_safe_repr(self):
        class CustomRepr(object):
            def __repr__(self):
                return 'hello'

        start = {'hello': 'world', 'custom': CustomRepr()}
        expected = copy.deepcopy(start)
        expected['custom'] = 'hello'
        self._assertSerialized(start, expected, safe_repr=False)

    def test_encode_with_custom_repr_safelist(self):
        class CustomRepr(object):
            def __repr__(self):
                return 'hello'

        start = {'hello': 'world', 'custom': CustomRepr()}
        expected = copy.deepcopy(start)
        expected['custom'] = 'hello'
        self._assertSerialized(start, expected, safelist=[CustomRepr])

    def test_encode_with_custom_repr_returns_bytes(self):
        class CustomRepr(object):
            def __repr__(self):
                return b'hello'

        start = {'hello': 'world', 'custom': CustomRepr()}

        serializable = SerializableTransform(safelist_types=[CustomRepr])
        result = transforms.transform(start, serializable)

        self.assertRegex(result['custom'], "<class '.*CustomRepr'>")

    def test_encode_with_custom_repr_returns_object(self):
        class CustomRepr(object):
            def __repr__(self):
                return {'hi': 'there'}

        start = {'hello': 'world', 'custom': CustomRepr()}

        serializable = SerializableTransform(safelist_types=[CustomRepr])
        result = transforms.transform(start, serializable)
        self.assertRegex(result['custom'], "<class '.*CustomRepr'>")

    def test_encode_with_custom_repr_returns_unicode(self):
        class CustomRepr(object):
            def __repr__(self):
                return SNOWMAN_UNICODE

        start = {'hello': 'world', 'custom': CustomRepr()}
        expected = copy.deepcopy(start)
        expected['custom'] = SNOWMAN_UNICODE
        self._assertSerialized(start, expected, safelist=[CustomRepr])

    def test_encode_with_bad_repr_doesnt_die(self):
        class CustomRepr(object):
            def __repr__(self):
                assert False

        start = {'hello': 'world', 'custom': CustomRepr()}
        serializable = SerializableTransform(safelist_types=[CustomRepr])
        result = transforms.transform(start, serializable)
        self.assertRegex(result['custom'], "<AssertionError.*CustomRepr.*>")

    def test_encode_with_bad_str_doesnt_die(self):

        class UnStringableException(Exception):
            def __str__(self):
                raise Exception('asdf')

        class CustomRepr(object):

            def __repr__(self):
                raise UnStringableException()

        start = {'hello': 'world', 'custom': CustomRepr()}
        serializable = SerializableTransform(safelist_types=[CustomRepr])
        result = transforms.transform(start, serializable)
        self.assertRegex(result['custom'], "<UnStringableException.*Exception.*str.*>")



import base64
import collections
import copy
from array import array

from collections.abc import Mapping

binary_type = bytes
integer_types = int
number_types = (float, int)
string_types = str
sequence_types = (Mapping, list, tuple, set, frozenset, array, collections.deque)


def force_lower(val):
    try:
        return val.lower()
    except:
        return str(val).lower()


def prefix_match(key, prefixes):
    if not key:
        return False

    for prefix in prefixes:
        if len(prefix) > len(key):
            continue

        if prefix == key[:len(prefix)]:
            return True

    return False


def key_in(key, canonicals):
    if not key:
        return False

    for c in canonicals:
        if key_match(key, c):
            return True

    return False


def key_depth(key, canonicals) -> int:
    if not key:
        return 0

    for c in canonicals:
        if key_match(key, c):
            return len(c)

    return 0


def key_match(key, canonical):
    if len(key) < len(canonical):
        return False

    for k, c in zip(key, canonical):
        if '*' == c:
            continue
        if c == k:
            continue
        return False

    return True


def reverse_list_of_lists(l, apply_each_fn=None):
    apply_each_fn = apply_each_fn or (lambda x: x)
    return [reversed([apply_each_fn(x) for x in inner]) for inner in l or []]


def build_key_matcher(prefixes_or_suffixes, type='prefix', case_sensitive=False):
    _prefixes = []

    if type == 'prefix':
        _iter = iter
    elif type == 'suffix':
        _iter = reversed
    else:
        raise ValueError('type must be either "prefix" or "suffix"')

    prefixes_or_suffixes = prefixes_or_suffixes or []
    for prefix in prefixes_or_suffixes:
        if case_sensitive:
            # Copy the list of lists
            _prefix = list(_iter(prefix))
        else:
            # Lowercase all of the elements
            _prefix = [force_lower(x) for x in _iter(prefix)]

        _prefixes.append(_prefix)

    def matcher(prefix_or_suffix):
        if case_sensitive:
            prefix = list(_iter(prefix_or_suffix))
        else:
            prefix = [force_lower(x) for x in _iter(prefix_or_suffix)]

        return prefix_match(prefix, _prefixes)

    return matcher


def is_builtin_type(obj):
    return obj.__class__.__module__ in ('__builtin__', 'builtins')


# http://www.xormedia.com/recursively-merge-dictionaries-in-python.html
def dict_merge(a, b, silence_errors=False):
    """
    Recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and bhave a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.
    """

    if not isinstance(b, dict):
        return b

    result = a
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v, silence_errors=silence_errors)
        else:
            try:
                result[k] = copy.deepcopy(v)
            except Exception as e:
                if not silence_errors:
                    raise e

                result[k] = '<Uncopyable obj:(%s)>' % (v,)

    return result


def circular_reference_label(data, ref_key=None):
    ref = '.'.join([str(x) for x in ref_key])
    return '<CircularReference type:(%s) ref:(%s)>' % (type(data).__name__, ref)


def float_nan_label(data):
    return '<NaN>'


def float_infinity_label(data):
    if data > 1:
        return '<Infinity>'
    else:
        return '<NegativeInfinity>'


def unencodable_object_label(data):
    return '<Unencodable type:(%s) base64:(%s)>' % (type(data).__name__,
                                                    base64.b64encode(data).decode('ascii'))


def undecodable_object_label(data):
    return '<Undecodable type:(%s) base64:(%s)>' % (type(data).__name__,
                                                    base64.b64encode(data).decode('ascii'))

try:
    from django.utils.functional import SimpleLazyObject
except ImportError:
    SimpleLazyObject = None


def defaultJSONEncode(o):
    if SimpleLazyObject and isinstance(o, SimpleLazyObject):
        if not o._wrapped:
            o._setup()
        return o._wrapped
    return repr(o) + " is not JSON serializable"



#!/usr/bin/env python

# This example uses Uvicorn package that must be installed. However, it can be
# replaced with any other ASGI-compliant server.
#
# NOTE: Python 3.6 requires aiocontextvars package to be installed.
#       Optional asynchronous reporting requires HTTPX package to be installed.
#
# Run: python app_logger.py

import logging

import rollbar
import uvicorn

from rollbar.contrib.starlette import LoggerMiddleware
from rollbar.logger import RollbarHandler
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse

# Initialize Rollbar SDK with your server-side ACCESS_TOKEN
rollbar.init(
    'ACCESS_TOKEN',
    environment='staging',
    handler='async',  # For asynchronous reporting use: default, async or httpx
)

# Set root logger to log DEBUG and above
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Report ERROR and above to Rollbar
rollbar_handler = RollbarHandler()
rollbar_handler.setLevel(logging.ERROR)

# Attach Rollbar handler to the root logger
logger.addHandler(rollbar_handler)

# Integrate Rollbar with Starlette application
app = Starlette()
app.add_middleware(LoggerMiddleware)  # should be added as the last middleware


# GET query params will be sent to Rollbar and available in the UI
# $ curl http://localhost:8888?param1=hello&param2=world
@app.route('/')
async def root(request):
    # Report log entries
    logger.critical('Critical message sent to Rollbar')
    logger.error('Error message sent to Rollbar')

    # Ignore log entries
    logger.warning('Warning message is not sent to Rollbar')
    logger.info('Info message is not sent to Rollbar')
    logger.debug('Debug message is not sent to Rollbar')

    return PlainTextResponse('hello world')


if __name__ == '__main__':
    uvicorn.run(app, host='localhost', port=8888)



#!/usr/bin/env python

# This example uses Uvicorn package that must be installed. However, it can be
# replaced with any other ASGI-compliant server.
#
# Optional asynchronous reporting requires HTTPX package to be installed.
#
# NOTE: This example requires FastAPI v0.41.0+ (see app_middleware.py for alternative).
#
# Run: python app.py

import fastapi
import rollbar
import uvicorn

from rollbar.contrib.fastapi import add_to as rollbar_add_to

# Initialize Rollbar SDK with your server-side ACCESS_TOKEN
rollbar.init(
    'ACCESS_TOKEN',
    environment='staging',
    handler='async',  # For asynchronous reporting use: default, async or httpx
    include_request_body=True,
)

# Integrate Rollbar with FastAPI application before adding routes to the app
app = fastapi.FastAPI()
rollbar_add_to(app)


# Verify application runs correctly
#
# $ curl http://localhost:8888
@app.get('/')
async def read_root():
    return {'hello': 'world'}


# Cause an uncaught exception to be sent to Rollbar
# GET query params will be sent to Rollbar and available in the UI
#
# $ curl http://localhost:8888/error?param1=hello&param2=world
async def localfunc(arg1, arg2, arg3):
    # Both local variables and function arguments will be sent to Rollbar
    # and available in the UI
    localvar = 'local variable'
    cause_error_with_local_variables


@app.get('/error')
async def read_error():
    await localfunc('func_arg1', 'func_arg2', 1)
    return {'result': "You shouldn't be seeing this"}


# Cause an uncaught exception to be sent to Rollbar
# POST request body will be sent to Rollbar and available in the UI
#
# curl http://localhost:8888/body -d '{"param1": "hello", "param2": "world"}'
@app.post('/body')
async def read_body():
    cause_error_with_body
    return {'result': "You shouldn't be seeing this"}


# Cause an uncaught exception to be sent to Rollbar
# POST form data will be sent to Rollbar and available in the UI
#
# curl http://localhost:8888/form -F 'param1=hello' -F 'param2=world'
@app.post('/form')
async def read_form():
    cause_error_with_form
    return {'result': "You shouldn't be seeing this"}


if __name__ == '__main__':
    uvicorn.run(app, host='localhost', port=8888)



def filter_rollbar_ignored_exceptions(exc_info, **kw):
    _, exc, _ = exc_info
    if getattr(exc, '_rollbar_ignore', False):
        return False

    return exc_info


def filter_by_level(target, **kw):
    if 'level' in kw and kw['level'] == 'ignored':
        return False

    return target



class Transform(object):
    depth_first = True

    def default(self, o, key=None):
        return o

    def transform_circular_reference(self, o, key=None, ref_key=None):
        # By default, we just perform a no-op for circular references.
        # Subclasses should implement this method to return whatever representation
        # for the circular reference they need.
        return self.default(o, key=key)

    def transform_tuple(self, o, key=None):
        return self.default(o, key=key)

    def transform_namedtuple(self, o, key=None):
        return self.default(o, key=key)

    def transform_list(self, o, key=None):
        return self.default(o, key=key)

    def transform_dict(self, o, key=None):
        return self.default(o, key=key)

    def transform_number(self, o, key=None):
        return self.default(o, key=key)

    def transform_bytes(self, o, key=None):
        return self.default(o, key=key)

    def transform_unicode(self, o, key=None):
        return self.default(o, key=key)

    def transform_boolean(self, o, key=None):
        return self.default(o, key=key)

    def transform_path(self, o, key=None):
        return self.default(str(o), key=key)

    def transform_custom(self, o, key=None):
        return self.default(o, key=key)
